"""Class for visualizing incident radiation (or irradiance) falling onto a mesh."""
from __future__ import division

try:  # first, assume we are in cPython and numpy is installed
    from typing import Tuple
    import numpy as np
except Exception:  # we are in IronPython or numpy is not installed
    np, Tuple = None, None

from ladybug_geometry.bounding import bounding_box
from ladybug_geometry.geometry3d import Mesh3D, Face3D
from ladybug.datatype.energyintensity import Radiation
from ladybug.datatype.energyflux import Irradiance
from ladybug.graphic import GraphicContainer
from ladybug.legend import LegendParameters
from ladybug.color import Colorset

from ..intersection import sky_intersection_matrix


class RadiationStudy(object):
    """Visualize the incident radiation (or irradiance) falling onto a study_mesh.

    Such studies of incident radiation can be used to approximate the energy that can
    be collected from photovoltaic or solar thermal systems. They are also useful
    for evaluating the impact of a building's orientation on both energy use and the
    size/cost of cooling systems. For studies of photovoltaic potential or building
    energy use impact, a sky matrix from EPW radiation should be used. For studies
    of cooling system size/cost, a sky matrix derived from the STAT file's clear sky
    radiation should be used.

    Not that no reflections of solar energy are included in the analysis performed by
    this class. Ground reflected irradiance is crudely accounted for by means of an
    emissive "ground hemisphere," which is like the sky dome hemisphere and is derived
    from the ground reflectance that is associated with the connected sky_matrix. This
    means that including geometry that represents the ground surface will effectively
    block such crude ground reflection.

    Args:
        sky_matrix: A SkyMatrix object, which describes the radiation coming
            from the various patches of the sky.
        study_mesh: A ladybug geometry Mesh3D, which represents the geometry on
            which the incident radiation is being studied.
        context_geometry: A list of ladybug geometry Face3D and/or Mesh3D that
            can block the view to the sky and ground.
        offset_distance: An optional number to offset the sensor points before
            the vectors are cast through the context_geometry. (Default: 0).
        by_vertex: A boolean to indicate whether the study should be run for
            each vertex of the study_mesh (True) or each face of the study
            mesh (False). (Default: False).
        sim_folder: An optional path to a folder where the simulation files will
            be written. If None, a temporary directory will be used. (Default: None).

    Properties:
        * sky_matrix
        * study_mesh
        * context_geometry
        * offset_distance
        * by_vertex
        * sim_folder
        * study_points
        * study_normals
        * intersection_matrix
        * radiation_values
        * irradiance_values
        * metadata
        * is_benefit
    """
    __slots__ = (
        '_metadata', '_is_benefit', '_sky_matrix', '_study_mesh', '_context_geometry',
        '_offset_distance', '_by_vertex', '_study_points', '_study_normals',
        '_sim_folder', '_intersection_matrix', '_radiation_values')

    def __init__(self, sky_matrix, study_mesh, context_geometry,
                 offset_distance=0, by_vertex=False, sim_folder=None):
        """Initialize RadiationStudy."""
        # set default values, which will be overwritten when the study is run
        self._offset_distance = float(offset_distance)
        self._by_vertex = bool(by_vertex)
        # set the key properties of the object
        self.sky_matrix = sky_matrix
        self.study_mesh = study_mesh
        self.context_geometry = context_geometry
        self.sim_folder = sim_folder
        # set default values, which will be overwritten when the study is run
        self._intersection_matrix = None
        self._radiation_values = None

    @property
    def sky_matrix(self):
        """Get or set a SkyMatrix object for the sky used in the study."""
        return self._sky_matrix

    @sky_matrix.setter
    def sky_matrix(self, value):
        self._sky_matrix = value
        self._metadata, direct, diffuse = value.data
        self._is_benefit = True if hasattr(value, 'benefit_matrix') and \
            value.benefit_matrix is not None else False
        self._radiation_values = None

    @property
    def study_mesh(self):
        """Get or set a SkyMatrix object for the sky used in the study."""
        return self._study_mesh

    @study_mesh.setter
    def study_mesh(self, value):
        assert isinstance(value, Mesh3D), 'Expected Mesh3D for RadiationStudy ' \
            'study_mesh. Got {}.'.format(type(value))
        self._study_mesh = value
        self._reset_points()

    @property
    def context_geometry(self):
        """Get or set a tuple of Face3D and Mesh3D for the geometry that can block sun.
        """
        return self._context_geometry

    @context_geometry.setter
    def context_geometry(self, value):
        if not isinstance(value, tuple):
            try:
                value = tuple(value)
            except (ValueError, TypeError):
                raise ValueError('Expected tuple or list for context_geometry. '
                                 'Got {}.'.format(type(value)))
        for geo in value:
            assert isinstance(geo, (Face3D, Mesh3D)), 'Expected Face3D or Mesh3D for ' \
                'RadiationStudy context_geometry. Got {}.'.format(type(geo))
        self._context_geometry = value
        self._intersection_matrix = None
        self._radiation_values = None

    @property
    def offset_distance(self):
        """Get or set a number for the offset distance used in the study."""
        return self._offset_distance

    @offset_distance.setter
    def offset_distance(self, value):
        assert isinstance(value, (float, int)), \
            'Expected number for offset_distance. Got {}.'.format(type(value))
        self._offset_distance = value
        self._reset_points()

    @property
    def by_vertex(self):
        """Get or set a boolean for whether the study should be run for each vertex."""
        return self._by_vertex

    @by_vertex.setter
    def by_vertex(self, value):
        self._by_vertex = bool(value)
        self._reset_points()

    @property
    def sim_folder(self):
        """Get or set text for the path where the simulation files are written."""
        return self._sim_folder

    @sim_folder.setter
    def sim_folder(self, value):
        if value is not None:
            assert isinstance(value, str), 'Expected file path string for sim_folder. ' \
                'Got {}.'.format(type(value))
        self._sim_folder = value

    @property
    def study_points(self):
        """Get a tuple of Point3Ds for the points of the study."""
        return self._study_points

    @property
    def study_normals(self):
        """Get a tuple of Vector3Ds for the normals of the study."""
        return self._study_normals

    @property
    def intersection_matrix(self):
        """Get a list of lists for the intersection matrix computed by the study."""
        if self._intersection_matrix is None:
            self._compute_intersection_matrix()
        return self._intersection_matrix

    @property
    def radiation_values(self):
        """Get a list of values for the radiation results of the study in kWh/m2."""
        if self._radiation_values is None:
            self.compute()
        return self._radiation_values

    @property
    def irradiance_values(self):
        """Get a list of values for the irradiance results of the study in W/m2."""
        factor = 1000 / self.sky_matrix.wea_duration \
            if hasattr(self.sky_matrix, 'wea_duration') else \
            1000 / (((self.metadata[3] - self.metadata[2]).total_seconds() / 3600) + 1)
        return [r * factor for r in self.radiation_values]

    @property
    def metadata(self):
        """Get a tuple of information about the metadata assigned to the study."""
        return self._metadata

    @property
    def is_benefit(self):
        """Get a boolean to note whether the sky matrix includes benefit information."""
        return self._is_benefit

    def total_radiation(self, conversion_to_meters=1):
        """Get a number for the total radiation of the study in kWh.

        Note that, when the study is run by_vertex, it is assumed that all vertices
        represent the same area.

        Args:
            conversion_to_meters: A number (typically less than 1) to note the
                conversion factor from the mesh area to square meters. This should
                be a conversion in square units and not just linear units. (Default: 1).
        """
        if self.by_vertex:
            full_area = self.study_mesh.area * conversion_to_meters
            total = sum(self.radiation_values) / full_area
        else:
            total = 0
            for rad, area in zip(self.radiation_values, self._study_mesh.face_areas):
                total += rad * area * conversion_to_meters
        return total

    def compute(self):
        """Compute the radiation values of the study.

        Note that this method is automatically called under the hood when accessing
        any results of the study and these results have not already been computed.
        So using this method is not necessary to correctly use this class. However,
        explicitly calling this method can help control when the time consuming
        part of the study runs, which is particularly helpful for larger studies.
        """
        # compute the intersection matrix
        if self._intersection_matrix is None:
            self._compute_intersection_matrix()
        # get the total radiation from the sky matrix
        mtx = self.sky_matrix.data
        if np is None:  # perform the calculation with float numbers
            sky_rad = [dir_rad + dif_rad for dir_rad, dif_rad in zip(mtx[1], mtx[2])]
            grd_val = (sum(sky_rad) / len(sky_rad)) * self.sky_matrix.ground_reflectance
            ground_rad = [grd_val] * len(sky_rad)
            all_rad = sky_rad + ground_rad
            self._radiation_values = [
                sum(r * w for r, w in zip(pt_rel, all_rad))
                for pt_rel in self._intersection_matrix
            ]
        else:  # perform the calculation with numpy matrices
            sky_rad = np.array(mtx[1]) + np.array(mtx[2])
            grd_val = (sky_rad.sum() / len(sky_rad)) * self.sky_matrix.ground_reflectance
            ground_rad = np.full(len(sky_rad), grd_val)
            all_rad = np.concatenate([sky_rad, ground_rad])
            self._radiation_values = np.dot(self._intersection_matrix, all_rad).tolist()

    def draw(self, legend_parameters=None, plot_irradiance=False):
        """Draw a colored study_mesh, compass, graphic/legend, and title.

        Args:
            legend_parameters: An optional LegendParameter object to change the display
                of the radiation study. If None, default legend parameters will be
                used. (Default: None).
            plot_irradiance: Boolean to note whether the results should be plotted
                with units of total Radiation (kWh/m2) [False] or with units of average
                Irradiance (W/m2) [True]. (Default: False).

        Returns:
            colored_mesh: A colored Mesh3D for the study results.
            graphic: A GraphicContainer for the colored mesh, indicating the
                legend and title location for the study.
            title: Text for the title of the study.
        """
        # get the radiation data
        if plot_irradiance:
            d_type, unit, title = Irradiance(), 'W/m2', 'Incident Irradiance'
            rad_data = self.irradiance_values
        else:
            d_type, unit, title = Radiation(), 'kWh/m2', 'Incident Radiation'
            rad_data = self.radiation_values
        if self.is_benefit:
            title = '{} Benefit/Harm'.format(title)

        # override the legend default min and max to make sense for the radiation study
        if legend_parameters is not None:
            assert isinstance(legend_parameters, LegendParameters), \
                'Expected LegendParameters. Got {}.'.format(type(legend_parameters))
            l_par = legend_parameters.duplicate()
        else:
            l_par = LegendParameters()
        if self.is_benefit:
            if l_par.min is None:
                l_par.min = min((min(rad_data), -max(rad_data)))
            if l_par.max is None:
                l_par.max = max((-min(rad_data), max(rad_data)))
            if l_par.are_colors_default:
                l_par.colors = reversed(Colorset.benefit_harm())
        else:
            if l_par.min is None:
                l_par.min = 0
            if l_par.max is None:
                l_par.max = max(rad_data)

        # create the mesh, graphic container, and title
        min_pt, max_pt = bounding_box((self.study_mesh,) + self.context_geometry)
        graphic = GraphicContainer(
            rad_data, min_pt, max_pt, l_par, d_type, unit)
        colored_mesh = self.study_mesh
        colored_mesh.colors = graphic.value_colors

        return colored_mesh, graphic, title

    def _reset_points(self):
        """Reset the study points and normals used in the study."""
        points = self._study_mesh.vertices if self._by_vertex else \
            self._study_mesh.face_centroids
        normals = self._study_mesh.vertex_normals if self._by_vertex else \
            self._study_mesh.face_normals
        if self._offset_distance != 0:
            points = tuple(
                pt.move(vec * self._offset_distance) for pt, vec in zip(points, normals))
        self._study_points = points
        self._study_normals = normals
        self._intersection_matrix = None
        self._radiation_values = None

    def _compute_intersection_matrix(self):
        """Compute intersection matrix."""
        self._intersection_matrix = sky_intersection_matrix(
            self.sky_matrix, self.study_points, self.study_normals,
            self.context_geometry, self.offset_distance, numericalize=True,
            sim_folder=self.sim_folder)

    def ToString(self):
        """Overwrite .NET ToString."""
        return self.__repr__()

    def __len__(self):
        return len(self.study_points)

    def __repr__(self):
        """Radiation Study object representation."""
        return 'RadiationStudy [{} values]'.format(len(self.study_points))



"""Class for visualizing the impact of radiation from different direction as a rose."""
from __future__ import division
import math

from ladybug_geometry.geometry2d.pointvector import Point2D
from ladybug_geometry.geometry3d.pointvector import Point3D, Vector3D
from ladybug_geometry.geometry3d.line import LineSegment3D
from ladybug_geometry.geometry3d.mesh import Mesh3D

from ladybug.datatype.energyintensity import Radiation
from ladybug.datatype.energyflux import Irradiance
from ladybug.viewsphere import view_sphere
from ladybug.compass import Compass
from ladybug.graphic import GraphicContainer
from ladybug.legend import LegendParameters
from ladybug.color import Colorset


class RadiationRose(object):
    """Visualize the impact of radiation from different direction as a rose.

    By default, the Radiation Rose depicts the amount of solar energy received
    by a vertical wall facing each of the directions of the compass rose. This
    is useful for understanding the radiation harm/benefit experienced by different
    building orientations or the orientations with the highest peak cooling load
    (for sky matrices of clear skies). The tilt_angle can be used to assess the
    solar energy falling on geometries that are not perfectly vertical, such
    as a tilted photovoltaic panel.

    Args:
        sky_matrix: A SkyMatrix object, which describes the radiation coming
            from the various patches of the sky.
        intersection_matrix: An optional lists of lists, which can be used to account
            for context shade surrounding the radiation rose. The matrix should
            have a length equal to the direction_count and begin from north moving
            clockwise. Each sub-list should consist of booleans and have a length
            equal to the number of sky patches times 2 (indicating sky patches and
            ground patches). True indicates that a certain patch is seen and False
            indicates that the match is blocked. If None, the radiation rose will
            be computed assuming no obstructions. (Default: None).
        direction_count: An integer greater than or equal to 3, which notes the number
            of arrows to be generated for the radiation rose. (Default: 36).
        tilt_angle: A number between 0 and 90 that sets the vertical tilt angle
            (aka. the altitude) for all of the directions. By default, the Radiation
            Rose depicts the amount of solar energy received by a vertical
            wall (tilt_angle=0). The tilt_angle can be changed to a specific
            value to assess the solar energy falling on geometries that are not
            perfectly vertical, such as a tilted photovoltaic panel. (Default: 0).
        legend_parameters: An optional LegendParameter object to change the display
            of the radiation rose. If None, default legend parameters will be
            used. (Default: None).
        plot_irradiance: Boolean to note whether the radiation rose should be plotted
            with units of total Radiation (kWh/m2) [False] or with units of average
            Irradiance (W/m2) [True]. (Default: False).
        center_point: A point for the center of the rose. (Default: (0, 0, 0)).
        radius: A number to set the radius of the radiation rose. (Default: 100).
        arrow_scale: A fractional number to note the scale of the radiation rose arrows
            in relation to the entire graphic. (Default: 1).

    Properties:
        * direction_count
        * tilt_angle
        * legend_parameters
        * plot_irradiance
        * center_point
        * radius
        * arrow_scale
        * north
        * direction_vectors
        * total_values
        * direct_values
        * diffuse_values
        * metadata
        * is_benefit
    """
    __slots__ = (
        '_north', '_direction_vectors', '_total_values', '_direct_values',
        '_diffuse_values', '_metadata', '_is_benefit', '_direction_count',
        '_tilt_angle', '_legend_parameters', '_plot_irradiance',
        '_center_point', '_radius', '_arrow_scale')

    def __init__(self, sky_matrix, intersection_matrix=None, direction_count=36,
                 tilt_angle=0, legend_parameters=None, plot_irradiance=False,
                 center_point=Point3D(0, 0, 0), radius=100, arrow_scale=1):
        """Initialize RadiationRose."""
        # deconstruct the sky matrix and derive key data from it
        metadata, direct, diffuse = sky_matrix.data
        self._metadata = metadata
        self._north = metadata[0]  # first item is the north angle
        self._plot_irradiance = bool(plot_irradiance)
        if plot_irradiance:
            factor = 1000 / sky_matrix.wea_duration \
                if hasattr(sky_matrix, 'wea_duration') else \
                1000 / (((metadata[3] - metadata[2]).total_seconds() / 3600) + 1)
            direct = tuple(v * factor for v in direct)
            diffuse = tuple(v * factor for v in diffuse)
        elif not isinstance(direct, tuple):
            direct, diffuse = tuple(direct), tuple(diffuse)

        # get the radiation coming from the ground
        dir_ground = ((sum(direct) / len(direct)) * metadata[1],) * len(direct)
        dif_ground = ((sum(diffuse) / len(diffuse)) * metadata[1],) * len(diffuse)
        all_dir = direct + dir_ground
        all_dif = diffuse + dif_ground

        # get the vectors and angles for each direction
        self._direction_count = int(direction_count)
        assert self._direction_count >= 3, 'RadiationRose direction_count must be ' \
            'greater than or equal to 3. Got {}.'.format(direction_count)
        assert isinstance(tilt_angle, (float, int)), 'Expected number for tilt_angle. ' \
            'Got {}.'.format(type(tilt_angle))
        assert 0 <= tilt_angle < 90, 'Rose tilt_angle must be between 0 and 90. ' \
            'Got {}.'.format(tilt_angle)
        self._tilt_angle = tilt_angle
        base_dir_vecs = view_sphere.horizontal_radial_vectors(self._direction_count)
        if self._tilt_angle != 0:
            x_axis = Vector3D(1, 0, 0)
            base_vec = Vector3D(0, 1, 0)
            horiz_angle = -2 * math.pi / self._direction_count
            vert_angle = math.radians(self._tilt_angle)
            dir_vecs = tuple(
                base_vec.rotate(x_axis, vert_angle).rotate_xy(horiz_angle * i)
                for i in range(self._direction_count))
        else:
            dir_vecs = base_dir_vecs
        patch_vecs = view_sphere.tregenza_sphere_vectors if len(direct) == 145 else \
            view_sphere.reinhart_sphere_vectors
        cos_angles = [[math.cos(v1.angle(v2)) for v2 in patch_vecs] for v1 in dir_vecs]
        if self._north != 0:
            na = math.radians(self._north)
            self._direction_vectors = tuple(vec.rotate_xy(na) for vec in base_dir_vecs)
        else:
            self._direction_vectors = base_dir_vecs

        # compute the radiation values for each direction
        point_relation = []
        if intersection_matrix is None:
            for cos_a in cos_angles:
                pt_rel = []
                for a in cos_a:
                    w = 0 if a < 0 else a
                    pt_rel.append(w)
                point_relation.append(pt_rel)
        else:
            for int_vals, cos_a in zip(intersection_matrix, cos_angles):
                pt_rel = []
                for iv, a in zip(int_vals, cos_a):
                    w = 0 if a < 0 or not iv else a
                    pt_rel.append(w)
                point_relation.append(pt_rel)
        total_res, direct_res, diff_res = [], [], []
        for pt_rel in point_relation:
            dir_v = sum(r * w for r, w in zip(pt_rel, all_dir))
            dif_v = sum(r * w for r, w in zip(pt_rel, all_dif))
            direct_res.append(dir_v)
            diff_res.append(dif_v)
            total_res.append(dir_v + dif_v)
        self._direct_values = tuple(direct_res)
        self._diffuse_values = tuple(diff_res)
        self._total_values = tuple(total_res)

        # override the legend default min and max to make sense for the radiation rose
        if legend_parameters is not None:
            assert isinstance(legend_parameters, LegendParameters), \
                'Expected LegendParameters. Got {}.'.format(type(legend_parameters))
            l_par = legend_parameters.duplicate()
        else:
            l_par = LegendParameters()
        if hasattr(sky_matrix, 'benefit_matrix') and \
                sky_matrix.benefit_matrix is not None:
            if l_par.min is None:
                l_par.min = min((min(self._total_values), -max(self._total_values)))
            if l_par.max is None:
                l_par.max = max((-min(self._total_values), max(self._total_values)))
            if l_par.are_colors_default:
                l_par.colors = reversed(Colorset.benefit_harm())
            self._is_benefit = True
        else:
            if l_par.min is None:
                l_par.min = 0
            if l_par.max is None:
                l_par.max = max(self._total_values)
            self._is_benefit = False
        self._legend_parameters = l_par

        # process the geometry parameters of the rose
        assert isinstance(center_point, Point3D), 'Expected Point3D for rose center. ' \
            'Got {}.'.format(type(center_point))
        self._center_point = center_point
        assert isinstance(radius, (float, int)), 'Expected number for radius. ' \
            'Got {}.'.format(type(radius))
        assert radius > 0, 'Rose radius must be greater than zero. ' \
            'Got {}.'.format(radius)
        self._radius = radius
        assert isinstance(arrow_scale, (float, int)), \
            'Expected number for arrow_scale. Got {}.'.format(type(arrow_scale))
        assert arrow_scale > 0, \
            'Rose arrow_scale must be greater than zero. Got {}.'.format(arrow_scale)
        self._arrow_scale = arrow_scale

    @property
    def direction_count(self):
        """Get the number of directions for the radiation rose."""
        return self._direction_count

    @property
    def tilt_angle(self):
        """Get the angle of the radiation rose vertical tilt in degrees."""
        return self._tilt_angle

    @property
    def legend_parameters(self):
        """Get the legend parameters assigned to this radiation rose object."""
        return self._legend_parameters

    @property
    def plot_irradiance(self):
        """Get a boolean for whether the rose values are for irradiance in (W/m2)."""
        return self._plot_irradiance

    @property
    def center_point(self):
        """Get a Point3D for the center of the radiation rose."""
        return self._center_point

    @property
    def radius(self):
        """Get a number for the radius of the radiation rose."""
        return self._radius

    @property
    def arrow_scale(self):
        """Get a number for the scale of arrows on the radiation rose."""
        return self._arrow_scale

    @property
    def north(self):
        """Get a number north direction."""
        return self._north

    @property
    def direction_vectors(self):
        """Get a list of vectors for each of the directions the rose is facing.

        All vectors are unit vectors.
        """
        return self._direction_vectors

    @property
    def total_values(self):
        """Get a tuple of values for the total radiation/irradiance of each direction."""
        return self._total_values

    @property
    def direct_values(self):
        """Get a tuple of values for the direct radiation/irradiance of each direction.
        """
        return self._direct_values

    @property
    def diffuse_values(self):
        """Get a tuple of values for the diffuse radiation/irradiance of each direction.
        """
        return self._diffuse_values

    @property
    def metadata(self):
        """Get a tuple of information about the metadata assigned to the radiation rose.
        """
        return self._metadata

    @property
    def is_benefit(self):
        """Boolean to note whether the sky matrix includes benefit information."""
        return self._is_benefit

    def draw(self, rad_type='total', center=None, max_rad=None):
        """Draw an arrow mesh, orientation lines, compass, graphic/legend, and title.

        Args:
            rad_type: Text for the type of radiation to use. Choose from total, direct,
                diffuse. (Default: total).
            center: A Point3D to override the center of the rose. This is useful
                when rendering all of the sky components together and one rose
                should not be on top of another. If None, the center
                point assigned to the object instance is used. (Default: None).
            max_rad: An optional number to set the level of radiation or irradiance
                associated with the full radius of the rose. If None, this is
                determined by the maximum level of radiation in the input data
                but a number can be specified here to fix this at a specific value.
                This is particularly useful when comparing different roses to one
                another. (Default: None).

        Returns:
            arrow_mesh: A colored Mesh3D for the rose arrows.
            orientation_lines: A list of LineSegment3D for each direction.
            compass: A ladybug Compass object for the rose.
            graphic: A GraphicContainer for the colored arrow mesh, indicating the
                legend and title location for the rose.
            rose_title: Text for the title of the rose.
        """
        # get the dome data to be plotted and the center point
        if rad_type.lower() == 'total':
            rad_data = self.total_values
        elif rad_type.lower() == 'direct':
            rad_data = self.direct_values
        elif rad_type.lower() == 'diffuse':
            rad_data = self.diffuse_values
        else:
            raise ValueError('Radiation type "{}" not recognized.'.format(rad_type))
        center = self.center_point if center is None else center

        # generate the mesh for the arrows of the rose
        ar_angle = math.pi / self.direction_count
        if max_rad is None:
            len_factor = self.radius / max(self.total_values) if not self.is_benefit \
                else self.radius / max((-min(self.total_values), max(self.total_values)))
        else:
            len_factor = self.radius / max_rad
        verts, faces, f_count = [center], [], 0
        for r_val, dir_vec in zip(rad_data, self.direction_vectors):
            # get key dimensions of the arrow
            arrow_length = abs(r_val) * len_factor
            head_dist = arrow_length * 0.8
            arrow_width = math.tan(ar_angle) * head_dist * 0.9 * self.arrow_scale
            h_vec = dir_vec * head_dist
            w_vec = dir_vec.rotate_xy(math.pi / 2) * arrow_width
            # create the mesh vertices
            right_pt = center.move(h_vec + w_vec)
            end_pt = center.move(dir_vec * arrow_length)
            left_pt = center.move(h_vec - w_vec)
            verts.extend((right_pt, end_pt, left_pt))
            faces.append((0, f_count + 1, f_count + 2, f_count + 3))
            f_count += 3
        arrow_mesh = Mesh3D(verts, faces)

        # generate the frequency lines
        low_cent = Point3D(center.x, center.y, center.z - 0.01)
        orientation_lines = [
            LineSegment3D(low_cent, vec * self.radius) for vec in self.direction_vectors]

        # output the rose visualization, including graphic and compass
        move_fac = self.radius * 1.15
        min_pt = center.move(Vector3D(-move_fac, -move_fac, 0))
        max_pt = center.move(Vector3D(move_fac, move_fac, 0))
        if self.plot_irradiance:
            d_type, unit, typ_str = Irradiance(), 'W/m2', 'Irradiance'
        else:
            d_type, unit, typ_str = Radiation(), 'kWh/m2', 'Radiation'
        graphic = GraphicContainer(
            rad_data, min_pt, max_pt, self.legend_parameters, d_type, unit)
        arrow_mesh.colors = graphic.value_colors
        compass = Compass(self.radius, Point2D(center.x, center.y), self.north)

        # construct a title from the metadata
        st, end = self.metadata[2], self.metadata[3]
        time_str = '{} - {}'.format(st, end) if st != end else st
        if self.is_benefit:
            typ_str = '{} Benefit/Harm'.format(typ_str)
        rose_title = '{} {}\n{}\n{}'.format(
            rad_type.title(), typ_str, time_str,
            '\n'.join([dat for dat in self.metadata[4:]]))

        return arrow_mesh, orientation_lines, compass, graphic, rose_title

    @staticmethod
    def radial_vectors(direction_count, tilt_angle=0):
        """Generate a list of radial vectors."""
        if tilt_angle != 0:
            x_axis = Vector3D(1, 0, 0)
            base_vec = Vector3D(0, 1, 0)
            horiz_angle = -2 * math.pi / direction_count
            vert_angle = math.radians(tilt_angle)
            return tuple(
                base_vec.rotate(x_axis, vert_angle).rotate_xy(horiz_angle * i)
                for i in range(direction_count))
        else:
            return view_sphere.horizontal_radial_vectors(direction_count)

    def ToString(self):
        """Overwrite .NET ToString."""
        return self.__repr__()

    def __len__(self):
        return self.direction_count

    def __getitem__(self, key):
        return self._total_values[key], self._direct_values[key], \
            self._diffuse_values[key]

    def __iter__(self):
        return zip(self._total_values, self._direct_values, self._diffuse_values)

    def __repr__(self):
        """Rose object representation."""
        return 'RadiationRose [{} directions]'.format(self.direction_count)



"""Class for visualizing the impact of radiation from different directions over a dome.
"""
from __future__ import division
import math

from ladybug_geometry.geometry2d.pointvector import Point2D
from ladybug_geometry.geometry3d.pointvector import Point3D, Vector3D
from ladybug_geometry.geometry3d.mesh import Mesh3D

from ladybug.datatype.energyintensity import Radiation
from ladybug.datatype.energyflux import Irradiance
from ladybug.viewsphere import view_sphere
from ladybug.compass import Compass
from ladybug.graphic import GraphicContainer
from ladybug.legend import LegendParameters
from ladybug.color import Colorset


class RadiationDome(object):
    """Visualize the radiation falling from different directions over a dome.

    The Radiation Dome depicts the amount of solar energy received by all directions
    over a dome. This is useful for understanding the optimal orientation of solar
    panels and how the performance of the panel will be impacted if it's orientation
    is off from the optimal position. It can also be used to identify the optimal
    wall orientation for passive solar heating when used with skies of
    radiation harm/benefit. When used with clear sky matrices, it can identify
    the orientations that result in the highest and lowest peak cooling load.

    Args:
        sky_matrix: A SkyMatrix object, which describes the radiation coming
            from the various patches of the sky.
        intersection_matrix: An optional lists of lists, which can be used to account
            for context shade surrounding the radiation dome. The matrix should
            have a length equal to the (azimuth_count * altitude_count) + 1 and begin
            from north moving clockwise, continuing up the dome with each revolution.
            The last vector refers to a perfectly vertical orientation.
            Each sub-list should consist of booleans and have a length equal to
            the number of sky patches times 2 (indicating sky patches followed by
            ground patches). True indicates that a certain patch is seen and False
            indicates that the match is blocked. If None, the radiation dome will
            be computed assuming no obstructions. (Default: None).
        azimuth_count: An integer greater than or equal to 3, which notes the number
            of horizontal orientations to be evaluated on the dome. (Default: 72).
        altitude_count: An integer greater than or equal to 3, which notes the number
            of vertical orientations to be evaluated on the dome. (Default: 18).
        legend_parameters: An optional LegendParameter object to change the display
            of the radiation dome. If None, default legend parameters will be
            used. (Default: None).
        plot_irradiance: Boolean to note whether the radiation dome should be plotted
            with units of total Radiation (kWh/m2) [False] or with units of average
            Irradiance (W/m2) [True]. (Default: False).
        center_point: A point for the center of the dome. (Default: (0, 0, 0)).
        radius: A number to set the radius of the radiation dome. (Default: 100).
        projection: Optional text for the name of a projection to use from the sky
            dome hemisphere to the 2D plane. If None, a 3D sky dome will be drawn
            instead of a 2D one. (Default: None) Choose from the following.

                * Orthographic
                * Stereographic

    Properties:
        * azimuth_count
        * altitude_count
        * legend_parameters
        * plot_irradiance
        * center_point
        * radius
        * projection
        * north
        * direction_vectors
        * dome_mesh
        * total_values
        * direct_values
        * diffuse_values
        * metadata
        * is_benefit
    """
    __slots__ = (
        '_north', '_metadata', '_is_benefit', '_direction_vectors', '_dome_mesh',
        '_total_values', '_direct_values', '_diffuse_values',
        '_azimuth_count', '_altitude_count', '_legend_parameters', '_plot_irradiance',
        '_center_point', '_radius', '_projection')
    PROJECTIONS = ('Orthographic', 'Stereographic')

    def __init__(self, sky_matrix, intersection_matrix=None, azimuth_count=72,
                 altitude_count=18, legend_parameters=None, plot_irradiance=False,
                 center_point=Point3D(0, 0, 0), radius=100, projection=None):
        """Initialize RadiationDome."""
        # deconstruct the sky matrix and derive key data from it
        metadata, direct, diffuse = sky_matrix.data
        self._metadata = metadata
        self._north = metadata[0]  # first item is the north angle
        self._plot_irradiance = bool(plot_irradiance)
        if plot_irradiance:
            factor = 1000 / sky_matrix.wea_duration \
                if hasattr(sky_matrix, 'wea_duration') else \
                1000 / (((metadata[3] - metadata[2]).total_seconds() / 3600) + 1)
            direct = tuple(v * factor for v in direct)
            diffuse = tuple(v * factor for v in diffuse)
        elif not isinstance(direct, tuple):
            direct, diffuse = tuple(direct), tuple(diffuse)

        # get the radiation coming from the ground
        dir_ground = ((sum(direct) / len(direct)) * metadata[1],) * len(direct)
        dif_ground = ((sum(diffuse) / len(diffuse)) * metadata[1],) * len(diffuse)
        all_dir = direct + dir_ground
        all_dif = diffuse + dif_ground

        # check the altitude and azimuth inputs
        self._azimuth_count = int(azimuth_count)
        assert self._azimuth_count >= 3, 'RadiationDome azimuth_count must be ' \
            'greater or equal to 3. Got {}.'.format(azimuth_count)
        self._altitude_count = int(altitude_count)
        assert self._altitude_count >= 3, 'RadiationDome altitude_count must be ' \
            'greater or equal to 3. Got {}.'.format(altitude_count)

        # get the vectors for each direction and compute their relation to the sky mtx
        dir_vecs = self.dome_vectors(self._azimuth_count, self._altitude_count)
        patch_vecs = view_sphere.tregenza_sphere_vectors if len(direct) == 145 else \
            view_sphere.reinhart_sphere_vectors
        cos_angles = [[math.cos(v1.angle(v2)) for v2 in patch_vecs] for v1 in dir_vecs]
        if self._north != 0:
            na = math.radians(self._north)
            self._direction_vectors = tuple(vec.rotate_xy(na) for vec in dir_vecs)
        else:
            self._direction_vectors = dir_vecs

        # compute the radiation values for each direction
        point_relation = []
        if intersection_matrix is None:
            for cos_a in cos_angles:
                pt_rel = []
                for a in cos_a:
                    w = 0 if a < 0 else a
                    pt_rel.append(w)
                point_relation.append(pt_rel)
        else:
            for int_vals, cos_a in zip(intersection_matrix, cos_angles):
                pt_rel = []
                for iv, a in zip(int_vals, cos_a):
                    w = 0 if a < 0 or not iv else a
                    pt_rel.append(w)
                point_relation.append(pt_rel)
        total_res, direct_res, diff_res = [], [], []
        for pt_rel in point_relation:
            dir_v = sum(r * w for r, w in zip(pt_rel, all_dir))
            dif_v = sum(r * w for r, w in zip(pt_rel, all_dif))
            direct_res.append(dir_v)
            diff_res.append(dif_v)
            total_res.append(dir_v + dif_v)
        self._direct_values = tuple(direct_res)
        self._diffuse_values = tuple(diff_res)
        self._total_values = tuple(total_res)

        # override the legend default min and max to make sense for the radiation dome
        if legend_parameters is not None:
            assert isinstance(legend_parameters, LegendParameters), \
                'Expected LegendParameters. Got {}.'.format(type(legend_parameters))
            l_par = legend_parameters.duplicate()
        else:
            l_par = LegendParameters()
        if hasattr(sky_matrix, 'benefit_matrix') and \
                sky_matrix.benefit_matrix is not None:
            if l_par.min is None:
                l_par.min = min((min(self._total_values), -max(self._total_values)))
            if l_par.max is None:
                l_par.max = max((-min(self._total_values), max(self._total_values)))
            if l_par.are_colors_default:
                l_par.colors = reversed(Colorset.benefit_harm())
            self._is_benefit = True
        else:
            if l_par.min is None:
                l_par.min = 0
            if l_par.max is None:
                l_par.max = max(self._total_values)
            self._is_benefit = False
        self._legend_parameters = l_par

        # process the geometry parameters of the dome
        assert isinstance(center_point, Point3D), 'Expected Point3D for dome center. ' \
            'Got {}.'.format(type(center_point))
        self._center_point = center_point
        assert isinstance(radius, (float, int)), 'Expected number for radius. ' \
            'Got {}.'.format(type(radius))
        assert radius > 0, 'Dome radius must be greater than zero. ' \
            'Got {}.'.format(radius)
        self._radius = radius
        if projection is not None:
            assert projection in self.PROJECTIONS, 'Projection "{}" is not recognized.' \
                ' Choose from: {}.'.format(projection, self.PROJECTIONS)
        self._projection = projection

        # use the direction vectors to create a mesh of the sky dome
        vertices = []
        for vec in self._direction_vectors:
            vertices.append(self.center_point.move(vec * self._radius))
        faces, pt_i, az_ct = [], 0, self._azimuth_count
        for row_count in range(self._altitude_count - 1):
            for _ in range(az_ct - 1):
                faces.append((pt_i, pt_i + 1, pt_i + az_ct + 1, pt_i + az_ct))
                pt_i += 1  # advance the number of vertices
            faces.append((pt_i, pt_i - az_ct + 1, pt_i + 1, pt_i + az_ct))
            pt_i += 1  # advance the number of vertices
        # add triangular faces to represent the last circular patch
        end_vert_i = len(vertices) - 1
        start_vert_i = len(vertices) - self._azimuth_count - 1
        for tr_i in range(0, self._azimuth_count - 1):
            faces.append((start_vert_i + tr_i, end_vert_i, start_vert_i + tr_i + 1))
        faces.append((end_vert_i - 1, end_vert_i, start_vert_i))
        self._dome_mesh = Mesh3D(vertices, faces)

    @property
    def azimuth_count(self):
        """Get the number of horizontal orientations for the radiation dome."""
        return self._azimuth_count

    @property
    def altitude_count(self):
        """Get the number of vertical directions for the radiation dome."""
        return self._altitude_count

    @property
    def legend_parameters(self):
        """Get the legend parameters assigned to this radiation dome object."""
        return self._legend_parameters

    @property
    def plot_irradiance(self):
        """Get a boolean for whether the dome values are for irradiance in (W/m2)."""
        return self._plot_irradiance

    @property
    def center_point(self):
        """Get a Point3D for the center of the radiation dome."""
        return self._center_point

    @property
    def radius(self):
        """Get a number for the radius of the radiation dome."""
        return self._radius

    @property
    def projection(self):
        """Get text for the projection of the dome."""
        return self._projection

    @property
    def north(self):
        """Get a number north direction."""
        return self._north

    @property
    def direction_vectors(self):
        """Get a list of vectors for each of the directions the dome is evaluating.

        All vectors are unit vectors.
        """
        return self._direction_vectors

    @property
    def dome_mesh(self):
        """Get a Mesh3D of the radiation dome with the center point and radius.

        This dome will be properly oriented to the north of the input sky matrix
        but it will not have any colors assigned to it.
        """
        return self._dome_mesh

    @property
    def total_values(self):
        """Get a tuple of values for the total radiation/irradiance of each direction."""
        return self._total_values

    @property
    def direct_values(self):
        """Get a tuple of values for the direct radiation/irradiance of each direction.
        """
        return self._direct_values

    @property
    def diffuse_values(self):
        """Get a tuple of values for the diffuse radiation/irradiance of each direction.
        """
        return self._diffuse_values

    @property
    def max_direction(self):
        """Get a the direction with the maximum total radiation/irradiance."""
        sort_v = [v for _, v in sorted(zip(self._total_values, self._direction_vectors))]
        return sort_v[-1]

    @property
    def max_point(self):
        """Get a point on the dome with the maximum total radiation/irradiance."""
        base_pt = self.center_point.move(self.max_direction * self.radius)
        if self.projection is not None:
            if self.projection.title() == 'Orthographic':
                base_pt2d = Compass.point3d_to_orthographic(base_pt)
                return Point3D(base_pt2d.x, base_pt2d.y, self.center_point.z)
            elif self.projection.title() == 'Stereographic':
                base_pt2d = Compass.point3d_to_stereographic(
                    base_pt, self.radius, self.center_point)
                return Point3D(base_pt2d.x, base_pt2d.y, self.center_point.z)
        return base_pt

    @property
    def max_info(self):
        """Get a text string with information about the maximum radiation/irradiance.

        This includes the altitude, azimuth, and radiation/irradiance value.
        """
        max_value = max(self.total_values)
        unit = 'W/m2' if self.plot_irradiance else 'kWh/m2'
        max_ind = self.total_values.index(max_value)
        max_az = (max_ind % self.azimuth_count) * (360 / self.azimuth_count)
        max_alt = math.floor(max_ind / self.azimuth_count) * (90 / self.altitude_count)
        return 'azimuth: {} deg\naltitude: {} deg\nvalue: {} {}'.format(
            int(max_az), int(max_alt), round(max_value, 1), unit)

    @property
    def metadata(self):
        """Get a tuple of information about the metadata assigned to the radiation dome.
        """
        return self._metadata

    @property
    def is_benefit(self):
        """Boolean to note whether the sky matrix includes benefit information."""
        return self._is_benefit

    def draw(self, rad_type='total', center=None):
        """Draw an dome mesh, compass, graphic/legend, and title.

        Args:
            rad_type: Text for the type of radiation to use. Choose from total, direct,
                diffuse. (Default: total).
            center: A Point3D to override the center of the dome. This is useful
                when rendering all of the sky components together and one dome
                should not be on top of another. If None, the center
                point assigned to the object instance is used. (Default: None).

        Returns:
            dome_mesh: A colored Mesh3D for the dome radiation.
            compass: A ladybug Compass object for the dome.
            graphic: A GraphicContainer for the colored arrow mesh, indicating the
                legend and title location for the dome.
            dome_title: Text for the title of the dome.
        """
        # get the dome data to be plotted and the center point
        if rad_type.lower() == 'total':
            rad_data = self.total_values
        elif rad_type.lower() == 'direct':
            rad_data = self.direct_values
        elif rad_type.lower() == 'diffuse':
            rad_data = self.diffuse_values
        else:
            raise ValueError('Radiation type "{}" not recognized.'.format(rad_type))

        if center is not None and center != self.center_point:
            center = center
            move_vec = center - self.center_point
            dome_mesh = self.dome_mesh.move(move_vec)
        else:
            center = self.center_point
            dome_mesh = self.dome_mesh

        # project the mesh if requested
        if self.projection is not None:
            if self.projection.title() == 'Orthographic':
                pts = (Compass.point3d_to_orthographic(pt) for pt in dome_mesh.vertices)
            elif self.projection.title() == 'Stereographic':
                pts = (Compass.point3d_to_stereographic(pt, self.radius, center)
                       for pt in dome_mesh.vertices)
            pts3d = tuple(Point3D(pt.x, pt.y, center.z) for pt in pts)
            dome_mesh = Mesh3D(pts3d, dome_mesh.faces)

        # output the dome visualization, including graphic and compass
        move_fac = self.radius * 1.15
        min_pt = center.move(Vector3D(-move_fac, -move_fac, 0))
        max_pt = center.move(Vector3D(move_fac, move_fac, 0))
        if self.plot_irradiance:
            d_type, unit, typ_str = Irradiance(), 'W/m2', 'Irradiance'
        else:
            d_type, unit, typ_str = Radiation(), 'kWh/m2', 'Radiation'
        graphic = GraphicContainer(
            rad_data, min_pt, max_pt, self.legend_parameters, d_type, unit)
        dome_mesh.colors = graphic.value_colors
        dome_compass = Compass(self.radius, Point2D(center.x, center.y), self.north)

        # construct a title from the metadata
        st, end = self.metadata[2], self.metadata[3]
        time_str = '{} - {}'.format(st, end) if st != end else st
        if self.is_benefit:
            typ_str = '{} Benefit/Harm'.format(typ_str)
        dome_title = '{} {}\n{}\n{}'.format(
            rad_type.title(), typ_str, time_str,
            '\n'.join([dat for dat in self.metadata[4:]]))

        return dome_mesh, dome_compass, graphic, dome_title

    @staticmethod
    def dome_vectors(azimuth_count, altitude_count):
        """Generate a list of vectors over the dome."""
        horiz_angle = -2 * math.pi / azimuth_count
        vert_angle = (math.pi / 2) / altitude_count
        dome_vecs = []
        for v in range(altitude_count):
            x_axis = Vector3D(1, 0, 0)
            base_vec = Vector3D(0, 1, 0)
            n_vec = base_vec.rotate(x_axis, vert_angle * v)
            for h in range(azimuth_count):
                dome_vecs.append(n_vec.rotate_xy(horiz_angle * h))
        dome_vecs.append(Vector3D(0, 0, 1))
        return dome_vecs

    def ToString(self):
        """Overwrite .NET ToString."""
        return self.__repr__()

    def __len__(self):
        return len(self._total_values)

    def __getitem__(self, key):
        return self._total_values[key], self._direct_values[key], \
            self._diffuse_values[key]

    def __iter__(self):
        return zip(self._total_values, self._direct_values, self._diffuse_values)

    def __repr__(self):
        """Radiation Dome object representation."""
        return 'RadiationDome [{} patches]'.format(len(self._total_values))



"""Class for visualizing sky matrices on a dome."""
from __future__ import division
import math

from ladybug_geometry.geometry2d.pointvector import Point2D
from ladybug_geometry.geometry3d.pointvector import Point3D, Vector3D
from ladybug_geometry.geometry3d.mesh import Mesh3D

from ladybug.datatype.energyintensity import Radiation
from ladybug.datatype.energyflux import Irradiance
from ladybug.viewsphere import view_sphere
from ladybug.compass import Compass
from ladybug.graphic import GraphicContainer
from ladybug.legend import LegendParameters
from ladybug.color import Colorset


class SkyDome(object):
    """Visualize a sky matrix as a colored dome, subdivided into patches.

    Args:
        sky_matrix: A SkyMatrix object, which describes the radiation coming
            from the various patches of the sky.
        legend_parameters: An optional LegendParameter object to change the display
            of the sky dome. If None, some default legend parameters will be
            used. (Default: None).
        plot_irradiance: Boolean to note whether the sky dome should be plotted with
            units of total Radiation (kWh/m2) [False] or with units of average
            Irradiance (W/m2) [True]. (Default: False).
        center_point: A point for the center of the dome. (Default: (0, 0, 0)).
        radius: A number to set the radius of the sky dome. (Default: 100).
        projection: Optional text for the name of a projection to use from the sky
            dome hemisphere to the 2D plane. If None, a 3D sky dome will be drawn
            instead of a 2D one. (Default: None) Choose from the following.

                * Orthographic
                * Stereographic

    Properties:
        * legend_parameters
        * plot_irradiance
        * center_point
        * radius
        * projection
        * north
        * patch_vectors
        * total_values
        * direct_values
        * diffuse_values
        * metadata
        * is_benefit
    """
    __slots__ = (
        '_north', '_total_values', '_direct_values', '_diffuse_values',
        '_metadata', '_is_benefit', '_legend_parameters', '_plot_irradiance',
        '_center_point', '_radius', '_projection')

    PROJECTIONS = ('Orthographic', 'Stereographic')

    def __init__(self, sky_matrix, legend_parameters=None, plot_irradiance=False,
                 center_point=Point3D(0, 0, 0), radius=100, projection=None):
        """Initialize SkyDome."""
        # deconstruct the sky matrix and derive key data from it
        metadata, direct, diffuse = sky_matrix.data
        self._north = metadata[0]  # first item is the north angle
        self._plot_irradiance = bool(plot_irradiance)
        if not plot_irradiance:
            self._direct_values = direct
            self._diffuse_values = diffuse
        else:
            factor = 1000 / sky_matrix.wea_duration \
                if hasattr(sky_matrix, 'wea_duration') else \
                1000 / (((metadata[3] - metadata[2]).total_seconds() / 3600) + 1)
            self._direct_values = tuple(v * factor for v in direct)
            self._diffuse_values = tuple(v * factor for v in diffuse)
        zip_obj = zip(self._direct_values, self._diffuse_values)
        self._total_values = tuple(dr + df for dr, df in zip_obj)
        self._metadata = metadata

        # override the legend default min and max to make sense for domes
        if legend_parameters is not None:
            assert isinstance(legend_parameters, LegendParameters), \
                'Expected LegendParameters. Got {}.'.format(type(legend_parameters))
            l_par = legend_parameters.duplicate()
        else:
            l_par = LegendParameters()
        if hasattr(sky_matrix, 'benefit_matrix') and \
                sky_matrix.benefit_matrix is not None:
            if l_par.min is None:
                l_par.min = min((min(self._total_values), -max(self._total_values)))
            if l_par.max is None:
                l_par.max = max((-min(self._total_values), max(self._total_values)))
            if l_par.are_colors_default:
                l_par.colors = reversed(Colorset.benefit_harm())
            self._is_benefit = True
        else:
            if l_par.min is None:
                l_par.min = 0
            if l_par.max is None:
                l_par.max = max(self._total_values)
            self._is_benefit = False
        self._legend_parameters = l_par

        # process the geometry parameters of the dome
        assert isinstance(center_point, Point3D), 'Expected Point3D for dome center. ' \
            'Got {}.'.format(type(center_point))
        self._center_point = center_point
        assert isinstance(radius, (float, int)), 'Expected number for radius. ' \
            'Got {}.'.format(type(radius))
        assert radius > 0, \
            'Dome radius must be greater than zero. Got {}.'.format(radius)
        self._radius = radius
        if projection is not None:
            assert projection in self.PROJECTIONS, 'Projection "{}" is not recognized.' \
                ' Choose from: {}.'.format(projection, self.PROJECTIONS)
        self._projection = projection

    @property
    def legend_parameters(self):
        """Get the legend parameters assigned to this sky dome object."""
        return self._legend_parameters

    @property
    def plot_irradiance(self):
        """Get a boolean for whether the sky dome values are for irradiance in (W/m2)."""
        return self._plot_irradiance

    @property
    def center_point(self):
        """Get a Point3D for the center of the dome."""
        return self._center_point

    @property
    def radius(self):
        """Get a number for the radius of the dome."""
        return self._radius

    @property
    def projection(self):
        """Get text for the projection of the dome."""
        return self._projection

    @property
    def north(self):
        """Get a number north direction."""
        return self._north

    @property
    def patch_vectors(self):
        """Get a list of vectors for each of the patches of the sky dome.

        All vectors are unit vectors and point from the center towards each
        of the patches. They can be used to construct visualizations of the
        rays used to perform radiation analysis.
        """
        return view_sphere.tregenza_dome_vectors if len(self._total_values) == 145 \
            else view_sphere.reinhart_dome_vectors

    @property
    def total_values(self):
        """Get a tuple of values for the total radiation/irradiance of each patch."""
        return self._total_values

    @property
    def direct_values(self):
        """Get a tuple of values for the direct radiation/irradiance of each patch."""
        return self._direct_values

    @property
    def diffuse_values(self):
        """Get a tuple of values for the diffuse radiation/irradiance of each patch."""
        return self._diffuse_values

    @property
    def metadata(self):
        """Get a tuple of information about the metadata assigned to the sky dome."""
        return self._metadata

    @property
    def is_benefit(self):
        """Boolean to note whether the sky matrix includes benefit information."""
        return self._is_benefit

    def draw(self, rad_type='total', center=None):
        """Draw a dome mesh, compass, graphic/legend, and title for the sky dome.

        Args:
            rad_type: Text for the type of radiation to use. Choose from total, direct,
                diffuse. (Default: total).
            center: A Point3D to override the center of the sky dome. This is useful
                when rendering all of the sky components together and one dome should
                not be on top of another. If None, the center point assigned to the
                object instance is used. (Default: None).

        Returns:
            dome_mesh: A colored Mesh3D for the dome.
            dome_compass: A ladybug Compass object for the dome.
            graphic: A GraphicContainer for the colored dome mesh, indicating the
                legend and title location for the dome.
            dome_title: Text for the title of the dome.
            values: A list of radiation values that align with the dome_mesh faces.
        """
        # get the dome data to be plotted
        if rad_type.lower() == 'total':
            dome_data = self.total_values
        elif rad_type.lower() == 'direct':
            dome_data = self.direct_values
        elif rad_type.lower() == 'diffuse':
            dome_data = self.diffuse_values
        else:
            raise ValueError('Radiation type "{}" not recognized.'.format(rad_type))

        # create the dome mesh and ensure patch values align with mesh faces
        if len(dome_data) == 145:  # tregenza sky
            dome_mesh = view_sphere.tregenza_dome_mesh_high_res.scale(self.radius)
            values = []  # high res dome has 3 x 3 faces per patch; we must convert
            tot_i = 0  # track the total number of patches converted
            for patch_i in view_sphere.TREGENZA_PATCHES_PER_ROW:
                row_vals = []
                for val in dome_data[tot_i:tot_i + patch_i]:
                    row_vals.extend([val] * 3)
                for i in range(3):
                    values.extend(row_vals)
                tot_i += patch_i
            values = values + [dome_data[-1]] * 18  # last patch has triangular faces
        else:  # reinhart sky
            dome_mesh = view_sphere.reinhart_dome_mesh.scale(self.radius)
            values = list(dome_data) + [dome_data[-1]] * 11  # triangular last patches

        # move and/or rotate the mesh as needed
        if self.north != 0:
            dome_mesh = dome_mesh.rotate_xy(math.radians(self.north), Point3D())
        center = self.center_point if center is None else center
        if center != Point3D():
            dome_mesh = dome_mesh.move(Vector3D(center.x, center.y, center.z))

        # project the mesh if requested
        if self.projection is not None:
            if self.projection.title() == 'Orthographic':
                pts = (Compass.point3d_to_orthographic(pt) for pt in dome_mesh.vertices)
            elif self.projection.title() == 'Stereographic':
                pts = (Compass.point3d_to_stereographic(pt, self.radius, center)
                       for pt in dome_mesh.vertices)
            pts3d = tuple(Point3D(pt.x, pt.y, center.z) for pt in pts)
            dome_mesh = Mesh3D(pts3d, dome_mesh.faces)

        # output the dome visualization, including graphic and compass
        move_fac = self.radius * 1.15
        min_pt = center.move(Vector3D(-move_fac, -move_fac, 0))
        max_pt = center.move(Vector3D(move_fac, move_fac, 0))
        if self.plot_irradiance:
            d_type, unit, typ_str = Irradiance(), 'W/m2', 'Irradiance'
        else:
            d_type, unit, typ_str = Radiation(), 'kWh/m2', 'Radiation'
        graphic = GraphicContainer(
            values, min_pt, max_pt, self.legend_parameters, d_type, unit)
        dome_mesh.colors = graphic.value_colors
        dome_compass = Compass(self.radius, Point2D(center.x, center.y), self.north)

        # construct a title from the metadata
        st, end = self.metadata[2], self.metadata[3]
        time_str = '{} - {}'.format(st, end) if st != end else st
        if self.is_benefit:
            typ_str = '{} Benefit/Harm'.format(typ_str)
        dome_title = '{} {}\n{}\n{}'.format(
            rad_type.title(), typ_str, time_str,
            '\n'.join([dat for dat in self.metadata[4:]]))

        return dome_mesh, dome_compass, graphic, dome_title, values

    def ToString(self):
        """Overwrite .NET ToString."""
        return self.__repr__()

    def __len__(self):
        return len(self._total_values)

    def __getitem__(self, key):
        return self._total_values[key], self._direct_values[key], \
            self._diffuse_values[key]

    def __iter__(self):
        return zip(self._total_values, self._direct_values, self._diffuse_values)

    def __repr__(self):
        """Sky Dome object representation."""
        return 'SkyDome [{} patches]'.format(len(self._total_values))



"""Get a matrix containing radiation values from each patch of a sky dome.

Creating this matrix is a necessary pre-step before doing incident radiation
analysis or generating a visualizations like a radiation rose.

This class uses Radiance's gendaymtx function to calculate the radiation
for each patch of the sky. Gendaymtx is written by Ian Ashdown and Greg Ward.
More information can be found in Radiance manual at:
http://www.radiance-online.org/learning/documentation/manual-pages/pdfs/gendaymtx.pdf
"""
from __future__ import division
import os
import subprocess
import sys

from ladybug.epw import EPW
from ladybug.wea import Wea
from ladybug.viewsphere import view_sphere
from ladybug.config import folders as lb_folders

from .config import folders

if folders.radbin_path is not None:
    GENDAYMTX_EXE = os.path.join(folders.radbin_path, 'gendaymtx.exe') if \
        os.name == 'nt' else os.path.join(folders.radbin_path, 'gendaymtx')
else:
    GENDAYMTX_EXE = None


class SkyMatrix(object):
    """A matrix containing radiation values from each patch of a sky dome.

    Args:
        wea: A Ladybug Wea object.
        north: A number between -360 and 360 for the counterclockwise difference
            between the North and the positive Y-axis in degrees. 90 is West and 270
            is East. (Default: 0).
        high_density: A Boolean to indicate whether the higher-density Reinhart
            sky matrix should be generated (True), which has roughly 4 times
            the sky patches as the (default) original Tregenza sky (False).
            Note that, while the Reinhart sky has a higher resolution and is
            more accurate, it will result in considerably longer calculation
            time for incident radiation studies. (Default: False).
        ground_reflectance: A number between 0 and 1 to note the average ground
            reflectance that is associated with the sky matrix. (Default: 0.2).

    Properties:
        * wea
        * north
        * high_density
        * ground_reflectance
        * benefit_matrix
        * folder
        * wea_duration
        * direct_values
        * diffuse_values
        * metadata
        * data
    """
    # constants for converting RGB values output by gendaymtx to broadband radiation
    PATCHES_PER_ROW = {
        1: view_sphere.TREGENZA_PATCHES_PER_ROW + (1,),
        2: view_sphere.REINHART_PATCHES_PER_ROW + (1,)
    }
    PATCH_ROW_COEFF = {
        1: view_sphere.TREGENZA_COEFFICIENTS,
        2: view_sphere.REINHART_COEFFICIENTS
    }
    LINE_BREAK = b'\n' if sys.version_info > (3, 0) else '\n'
    SEPARATOR = b' ' if sys.version_info > (3, 0) else ' '

    __slots__ = (
        '_wea', '_north', '_high_density', '_ground_reflectance', '_folder',
        '_direct_values', '_diffuse_values', '_benefit_matrix', '_metadata')

    def __init__(self, wea, north=0, high_density=False, ground_reflectance=0.2):
        """Initialize SkyMatrix."""
        self.wea = wea
        self.north = north
        self.high_density = high_density
        self.ground_reflectance = ground_reflectance
        self.benefit_matrix = None
        self.folder = None

    @classmethod
    def from_components(
            cls, location, direct_normal_irradiance, diffuse_horizontal_irradiance,
            hoys=None, north=0, high_density=False, ground_reflectance=0.2):
        """Create a SkyMatrix from individual solar irradiance components.

        Args:
            location: Ladybug location object.
            direct_normal_irradiance: A HourlyContinuousCollection or a
                HourlyDiscontinuousCollection for direct normal irradiance. The
                collection must be aligned with the diffuse_horizontal_irradiance.
            diffuse_horizontal_irradiance: A HourlyContinuousCollection or a
                HourlyDiscontinuousCollection for diffuse horizontal irradiance, The
                collection must be aligned with the direct_normal_irradiance.
            hoys: A list of numbers between 0 and 8760 that represent the hours of the
                year for which to generate the sky matrix. If None, the matrix will
                be for the entire year. (Default: None).
            north: A number between -360 and 360 for the counterclockwise difference
                between the North and the positive Y-axis in degrees. 90 is West
                and 270 is East. (Default: 0).
            high_density: A Boolean to indicate whether the higher-density Reinhart
                sky matrix should be generated (True), which has roughly 4 times
                the sky patches as the (default) original Tregenza sky (False).
                Note that, while the Reinhart sky has a higher resolution and is
                more accurate, it will result in considerably longer calculation
                time for incident radiation studies. (Default: False).
            ground_reflectance: A number between 0 and 1 to note the average ground
                reflectance that is associated with the sky matrix. (Default: 0.2).
        """
        # filter the radiation by hoys if they are input
        if hoys is not None and len(hoys) != 0:
            dni = direct_normal_irradiance.filter_by_hoys(hoys)
            dhi = diffuse_horizontal_irradiance.filter_by_hoys(hoys)
        else:
            dni, dhi = direct_normal_irradiance, diffuse_horizontal_irradiance
        # create the wea and return the SkyMatrix
        wea = Wea(location, dni, dhi)
        return cls(wea, north, high_density, ground_reflectance)

    @classmethod
    def from_components_benefit(
            cls, location, direct_normal_irradiance, diffuse_horizontal_irradiance,
            temperature, balance_temperature=15, balance_offset=2,
            hoys=None, north=0, high_density=False, ground_reflectance=0.2):
        """Create a SkyMatrix representing benefit/harm based on temperature data.

        Args:
            location: Ladybug location object.
            direct_normal_irradiance: A HourlyContinuousCollection or a
                HourlyDiscontinuousCollection for direct normal irradiance. The
                collection must be aligned with the diffuse_horizontal_irradiance.
            diffuse_horizontal_irradiance: A HourlyContinuousCollection or a
                HourlyDiscontinuousCollection for diffuse horizontal irradiance. The
                collection must be aligned with the direct_normal_irradiance.
            temperature: A HourlyContinuousCollection or a HourlyDiscontinuousCollection
                for temperature, which will be used to establish whether radiation
                is desired or not for each time step. The collection must be aligned
                with the irradiance inputs.
            balance_temperature: The temperature in Celsius between which radiation
                switches from being a benefit to a harm. Typical residential buildings
                have balance temperatures as high as 18C and commercial buildings tend
                to have lower values around 12C. (Default 15C).
            balance_offset: The temperature offset from the balance temperature
                in Celsius where radiation is neither harmful nor helpful. (Default: 2).
            hoys: A list of numbers between 0 and 8760 that represent the hours of the
                year for which to generate the sky matrix. If None, the matrix will
                be for the entire year. (Default: None).
            north: A number between -360 and 360 for the counterclockwise difference
                between the North and the positive Y-axis in degrees. 90 is West
                and 270 is East. (Default: 0).
            high_density: A Boolean to indicate whether the higher-density Reinhart
                sky matrix should be generated (True), which has roughly 4 times
                the sky patches as the (default) original Tregenza sky (False).
                Note that, while the Reinhart sky has a higher resolution and is
                more accurate, it will result in considerably longer calculation
                time for incident radiation studies. (Default: False).
            ground_reflectance: A number between 0 and 1 to note the average ground
                reflectance that is associated with the sky matrix. (Default: 0.2).
        """
        # filter the radiation by hoys if they are input
        if hoys is not None and len(hoys) != 0:
            dni = direct_normal_irradiance.filter_by_hoys(hoys)
            dhi = diffuse_horizontal_irradiance.filter_by_hoys(hoys)
            temperature = temperature.filter_by_hoys(hoys)
        else:
            dni, dhi = direct_normal_irradiance, diffuse_horizontal_irradiance
        # create the benefit matrix
        t_low = balance_temperature - balance_offset
        t_high = balance_temperature + balance_offset
        benefit_mtx = []
        for t in temperature:
            if t < t_low:  # low temperatures mean radiation is beneficial
                benefit_mtx.append(True)
            elif t > t_high:  # high temperatures mean radiation is harmful
                benefit_mtx.append(False)
            else:  # the temperature is neutral, meaning radiation has no effect
                benefit_mtx.append(None)
        # create the wea and return the SkyMatrix
        wea = Wea(location, dni, dhi)
        sky_mtx = cls(wea, north, high_density, ground_reflectance)
        sky_mtx.benefit_matrix = benefit_mtx
        return sky_mtx

    @classmethod
    def from_epw(cls, epw_file, hoys=None, north=0, high_density=False,
                 ground_reflectance=0.2):
        """Create a SkyMatrix using the solar irradiance values in an epw file.

        Args:
            epw_file: Full path to epw weather file.
            hoys: A list of numbers between 0 and 8760 that represent the hours of the
                year for which to generate the sky matrix. If None, the matrix will
                be for the entire year. (Default: None).
            north: A number between -360 and 360 for the counterclockwise difference
                between the North and the positive Y-axis in degrees. 90 is West
                and 270 is East. (Default: 0).
            high_density: A Boolean to indicate whether the higher-density Reinhart
                sky matrix should be generated (True), which has roughly 4 times
                the sky patches as the (default) original Tregenza sky (False).
                Note that, while the Reinhart sky has a higher resolution and is
                more accurate, it will result in considerably longer calculation
                time for incident radiation studies. (Default: False).
            ground_reflectance: A number between 0 and 1 to note the average ground
                reflectance that is associated with the sky matrix. (Default: 0.2).
        """
        epw = EPW(epw_file)
        return cls.from_components(
            epw.location, epw.direct_normal_radiation, epw.diffuse_horizontal_radiation,
            hoys, north, high_density, ground_reflectance)

    @classmethod
    def from_epw_benefit(cls, epw_file, balance_temperature=15, balance_offset=2,
                         hoys=None, north=0, high_density=False, ground_reflectance=0.2):
        """Create a SkyMatrix using the solar irradiance values in an epw file.

        Args:
            epw_file: Full path to epw weather file.
            balance_temperature: The temperature in Celsius between which radiation
                switches from being a benefit to a harm. Typical residential buildings
                have balance temperatures as high as 18C and commercial buildings tend
                to have lower values around 12C. (Default 15C).
            balance_offset: The temperature offset from the balance temperature
                in Celsius where radiation is neither harmful nor helpful. (Default: 2).
            hoys: A list of numbers between 0 and 8760 that represent the hours of the
                year for which to generate the sky matrix. If None, the matrix will
                be for the entire year. (Default: None).
            north: A number between -360 and 360 for the counterclockwise difference
                between the North and the positive Y-axis in degrees. 90 is West
                and 270 is East. (Default: 0).
            high_density: A Boolean to indicate whether the higher-density Reinhart
                sky matrix should be generated (True), which has roughly 4 times
                the sky patches as the (default) original Tregenza sky (False).
                Note that, while the Reinhart sky has a higher resolution and is
                more accurate, it will result in considerably longer calculation
                time for incident radiation studies. (Default: False).
            ground_reflectance: A number between 0 and 1 to note the average ground
                reflectance that is associated with the sky matrix. (Default: 0.2).
        """
        epw = EPW(epw_file)
        return cls.from_components_benefit(
            epw.location, epw.direct_normal_radiation, epw.diffuse_horizontal_radiation,
            epw.dry_bulb_temperature, balance_temperature, balance_offset,
            hoys, north, high_density, ground_reflectance)

    @classmethod
    def from_stat(cls, stat_file, hoys=None, north=0, high_density=False,
                  ground_reflectance=0.2):
        """Create a ASHRAE Revised Clear SkyMatrix using the data in .stat file.

        The .stat file must have monthly sky optical depths within it in order to
        create a Wea this way.

        Args:
            stat_file: Full path to a .stat file.
            hoys: A list of numbers between 0 and 8760 that represent the hours of the
                year for which to generate the sky matrix. If None, the matrix will
                be for the entire year. (Default: None).
            north: A number between -360 and 360 for the counterclockwise difference
                between the North and the positive Y-axis in degrees. 90 is West
                and 270 is East. (Default: 0).
            high_density: A Boolean to indicate whether the higher-density Reinhart
                sky matrix should be generated (True), which has roughly 4 times
                the sky patches as the (default) original Tregenza sky (False).
                Note that, while the Reinhart sky has a higher resolution and is
                more accurate, it will result in considerably longer calculation
                time for incident radiation studies. (Default: False).
            ground_reflectance: A number between 0 and 1 to note the average ground
                reflectance that is associated with the sky matrix. (Default: 0.2).
        """
        wea = Wea.from_stat_file(stat_file)
        # filter the radiation by hoys if they are input
        if hoys is not None and len(hoys) != 0:
            wea = wea.filter_by_hoys(hoys)
        return cls(wea, north, high_density, ground_reflectance)

    @classmethod
    def from_ashrae_clear_sky(cls, location, sky_clearness=1, hoys=None, north=0,
                              high_density=False, ground_reflectance=0.2):
        """Create an original ASHRAE Clear SkyMatrix using a location and sky clearness.

        The original ASHRAE Clear Sky is intended to determine peak solar load
        and sizing parameters for HVAC systems.  It is not the sky model
        currently recommended by ASHRAE since it usually overestimates the
        amount of solar irradiance in comparison to the newer ASHRAE Revised
        Clear Sky ("Tau Model"). However, the original model here is still
        useful for cases where monthly optical depth values are not known. For
        more information on the ASHRAE Clear Sky model, see the EnergyPlus
        Engineering Reference:
        https://bigladdersoftware.com/epx/docs/8-9/engineering-reference/climate-calculations.html

        Args:
            location: Ladybug location object.
            sky_clearness: A factor that will be multiplied by the output of
                the model. This is to help account for locations where clear,
                dry skies predominate (e.g., at high elevations) or,
                conversely, where hazy and humid conditions are frequent. See
                Threlkeld and Jordan (1958) for recommended values. Typical
                values range from 0.95 to 1.05 and are usually never more
                than 1.2. Default is set to 1.0.
            hoys: A list of numbers between 0 and 8760 that represent the hours of the
                year for which to generate the sky matrix. If None, the matrix will
                be for the entire year. (Default: None).
            north: A number between -360 and 360 for the counterclockwise difference
                between the North and the positive Y-axis in degrees. 90 is West
                and 270 is East. (Default: 0).
            high_density: A Boolean to indicate whether the higher-density Reinhart
                sky matrix should be generated (True), which has roughly 4 times
                the sky patches as the (default) original Tregenza sky (False).
                Note that, while the Reinhart sky has a higher resolution and is
                more accurate, it will result in considerably longer calculation
                time for incident radiation studies. (Default: False).
            ground_reflectance: A number between 0 and 1 to note the average ground
                reflectance that is associated with the sky matrix. (Default: 0.2).
        """
        wea = Wea.from_ashrae_clear_sky(location, sky_clearness)
        # filter the radiation by hoys if they are input
        if hoys is not None and len(hoys) != 0:
            wea = wea.filter_by_hoys(hoys)
        return cls(wea, north, high_density, ground_reflectance)

    @property
    def wea(self):
        """Get or set a Wea object for the sky matrix."""
        return self._wea

    @wea.setter
    def wea(self, value):
        assert isinstance(value, Wea), \
            'wea must be from type Wea not {}'.format(type(value))
        self._wea = value
        self._direct_values = None
        self._diffuse_values = None
        self._metadata = None

    @property
    def north(self):
        """Get or set a number north direction.

        A number between -360 and 360 for the counterclockwise difference between
        the North and the positive Y-axis in degrees. 90 is West and 270 is East.
        """
        return self._north

    @north.setter
    def north(self, value):
        assert isinstance(value, (float, int)), 'Expected number for ' \
            'SkyMatrix north. Got {}.'.format(type(value))
        assert -360 <= value <= 360, 'SkyMatrix north must be between 0 and 360. ' \
            'Got {}.'.format(value)
        self._north = value

    @property
    def high_density(self):
        """Get or set a boolean for whether the sky is a higher-density Reinhart matrix.
        """
        return self._high_density

    @high_density.setter
    def high_density(self, value):
        self._high_density = bool(value)
        self._direct_values = None
        self._diffuse_values = None
        self._metadata = None

    @property
    def ground_reflectance(self):
        """Get or set a number between 0 and 1 to note the average ground reflectance.
        """
        return self._ground_reflectance

    @ground_reflectance.setter
    def ground_reflectance(self, value):
        assert isinstance(value, (float, int)), 'Expected number for ' \
            'SkyMatrix ground_reflectance. Got {}.'.format(type(value))
        assert 0 <= value <= 1, 'SkyMatrix ground_reflectance must be between 0 and 1.' \
            ' Got {}.'.format(value)
        self._ground_reflectance = value

    @property
    def benefit_matrix(self):
        """Get or set list of True/False values for whether Wea datetimes are beneficial.

        This list must have a length that matches the Wea so that each datetime
        can be matched with a benefit/harm value. True (beneficial) values will
        contribute positively to the value of each sky patch while False (harmful)
        values will contribute negatively. A value of None in the list indicates
        that the hour has neither a benefit or a harmful effect.

        This is None by default, indicating that all radiation values contribute
        positively to each sky patch.
        """
        return self._benefit_matrix

    @benefit_matrix.setter
    def benefit_matrix(self, value):
        if value is not None:
            assert isinstance(value, (list, tuple)), \
                'SkyMatrix.benefit_matrix must be a list. Not {}'.format(type(value))
            assert len(value) == len(self.wea), 'Length of SkyMatrix.benefit_matrix ' \
                '[{}] must equal the length of the sky matrix Wea [{}].'.format(
                    len(value), len(self.wea))
        self._benefit_matrix = value
        self._direct_values = None
        self._diffuse_values = None
        self._metadata = None

    @property
    def folder(self):
        """Get or set the folder in which the Radiance commands are executed.

        If None, it will be written to Ladybug's default EPW folder.
        """
        if self._folder is None:
            return os.path.join(lb_folders.default_epw_folder, 'sky_matrices')
        return self._folder

    @folder.setter
    def folder(self, value):
        if value is not None:
            assert os.path.isdir(value), 'Path for ' \
                'SkyMatrix folder does not exist: {}'.format(type(value))
        self._folder = value

    @property
    def wea_duration(self):
        """Get the duration of the Wea in hours.

        This is useful for converting the radiation values of the sky patches (kWh/m2)
        into irradiance (W/m2).
        """
        return len(self.wea) / self.wea.timestep

    @property
    def direct_values(self):
        """Get the direct radiation values for each of the sky patches."""
        if self._direct_values is None:
            self.compute_sky()
        return self._direct_values

    @property
    def diffuse_values(self):
        """Get the diffuse radiation values for each of the sky patches."""
        if self._diffuse_values is None:
            self.compute_sky()
        return self._diffuse_values

    @property
    def metadata(self):
        """Get a list of metadata associated with the sky matrix."""
        if self._metadata is None:
            self.compute_sky()
        return self._metadata

    @property
    def data(self):
        """Get a matrix of all data associated with the sky matrix.

        The first list contains metadata, followed by direct values and then
        diffuse values.
        """
        if self._metadata is None:
            self.compute_sky()
        return (self._metadata, self._direct_values, self._diffuse_values)

    def compute_sky(self):
        """Compute the values of the sky matrix."""
        # extract metadata needed for all calculations
        wea_duration = len(self.wea) / self.wea.timestep
        metd = self.wea.direct_normal_irradiance.header.metadata
        wea_basename = metd['city'].replace(' ', '_') if 'city' in metd else 'unnamed'

        # if there's a benefit matrix, split the Wea into two files
        if self.benefit_matrix is not None:
            dir_vals1, dif_vals1, dir_vals2, dif_vals2 = [], [], [], []
            zip_obj = zip(
                self.wea.direct_normal_irradiance,
                self.wea.diffuse_horizontal_irradiance,
                self.benefit_matrix
            )
            for dir_v, dif_v, ben in zip_obj:
                if ben:  # time contributes positively
                    dir_vals1.append(dir_v)
                    dif_vals1.append(dif_v)
                    dir_vals2.append(0)
                    dif_vals2.append(0)
                elif ben is None:  # time is neither positive nor negative
                    dir_vals1.append(0)
                    dif_vals1.append(0)
                    dir_vals2.append(0)
                    dif_vals2.append(0)
                else:  # time contributes negatively
                    dir_vals2.append(dir_v)
                    dif_vals2.append(dif_v)
                    dir_vals1.append(0)
                    dif_vals1.append(0)
            # create the first Wea object
            dir_data1 = self.wea.direct_normal_irradiance.duplicate()
            dif_data1 = self.wea.diffuse_horizontal_irradiance.duplicate()
            dir_data1.values = dir_vals1
            dif_data1.values = dif_vals1
            wea_obj1 = Wea(self.wea.location, dir_data1, dif_data1)
            # create the second Wea object
            dir_data2 = self.wea.direct_normal_irradiance.duplicate()
            dif_data2 = self.wea.diffuse_horizontal_irradiance.duplicate()
            dir_data2.values = dir_vals2
            dif_data2.values = dif_vals2
            wea_obj2 = Wea(self.wea.location, dir_data2, dif_data2)
            # write the Wea files
            wea_path1 = os.path.join(self.folder, '{}_benefit'.format(wea_basename))
            wea_path2 = os.path.join(self.folder, '{}_harm'.format(wea_basename))
            wea_file = wea_obj1.write(wea_path1)
            wea_file2 = wea_obj2.write(wea_path2)
        else:  # otherwise, write the Wea to the folder
            wea_path = os.path.join(self.folder, wea_basename)
            wea_file = self.wea.write(wea_path)
            wea_file2 = None

        # execute the Radiance gendaymtx command
        dir_vals, dif_vals = self._run_gendaymtx(wea_file, wea_duration)
        # if there's a second wea, then use it to compute radiation harm
        if wea_file2 is not None:
            dir_vals2, dif_vals2 = self._run_gendaymtx(wea_file2, wea_duration)
            self._direct_values = tuple(db - dh for db, dh in zip(dir_vals, dir_vals2))
            self._diffuse_values = tuple(db - dh for db, dh in zip(dif_vals, dif_vals2))
        else:
            self._direct_values = dir_vals
            self._diffuse_values = dif_vals

        # collect sky metadata like the north, which will be used by other operations
        metadata = [self.north, self.ground_reflectance]
        dts = self.wea.direct_normal_irradiance.datetimes
        metadata.extend([dts[0], dts[-1]])
        for key, val in self.wea.direct_normal_irradiance.header.metadata.items():
            metadata.append('{} : {}'.format(key, val))
        self._metadata = tuple(metadata)

    def _run_gendaymtx(self, wea_file, wea_duration):
        """Run a Wea file through gendaymtx and get direct and diffuse radiation.

        Args:
            wea_file: Path to a Wea file to be run through gendaymtx.
            wea_duration: Number for the duration of the Wea in hours. This is used
                to convert between the average value output by the command and the
                cumulative value that is needed for all ladybug analyses.
        """
        assert GENDAYMTX_EXE is not None, 'No Radiance installation was found.'
        density = 2 if self.high_density else 1
        use_shell = True if os.name == 'nt' else False
        # command for direct patches
        cmds = [GENDAYMTX_EXE, '-m', str(density), '-d', '-O1', '-A', wea_file]
        process = subprocess.Popen(cmds, stdout=subprocess.PIPE, shell=use_shell)
        stdout = process.communicate()
        dir_data_str = stdout[0]
        # command for diffuse patches
        cmds = [GENDAYMTX_EXE, '-m', str(density), '-s', '-O1', '-A', wea_file]
        process = subprocess.Popen(cmds, stdout=subprocess.PIPE, shell=use_shell)
        stdout = process.communicate()
        diff_data_str = stdout[0]
        # parse the data into a single matrix
        dir_vals = self._parse_mtx_data(dir_data_str, wea_duration, density)
        diff_vals = self._parse_mtx_data(diff_data_str, wea_duration, density)
        return dir_vals, diff_vals

    def _parse_mtx_data(self, data_str, wea_duration, sky_density=1):
        """Parse a string of Radiance gendaymtx data to a list of radiation-per-patch.

        This function handles the removing of the header and the conversion of the
        RGB irradiance=per-steradian values to broadband radiation. It also removes
        the first patch, which is the ground and is not used by Ladybug.

        Args:
            data_str: The string that has been output by gendaymtx to stdout.
            wea_duration: Number for the duration of the Wea in hours. This is used
                to convert between the average value output by the command and the
                cumulative value that is needed for all ladybug analyses.
            sky_density: Integer (either 1 or 2) for the density.
        """
        # split lines and remove the header, ground patch and last line break
        data_lines = data_str.split(self.LINE_BREAK)
        patch_lines = data_lines[9:-1]

        # loop through the rows and convert the radiation RGB values
        broadband_irr = []
        patch_counter = 0
        for i, row_patch_count in enumerate(self.PATCHES_PER_ROW[sky_density]):
            row_slice = patch_lines[patch_counter:patch_counter + row_patch_count]
            irr_vals = (self._broadband_radiation(row, i, wea_duration, sky_density)
                        for row in row_slice)
            broadband_irr.extend(irr_vals)
            patch_counter += row_patch_count
        return tuple(broadband_irr)

    def _broadband_radiation(
            self, patch_row_str, row_number, wea_duration, sky_density=1):
        """Parse a row of gendaymtx RGB patch data in W/sr/m2 to radiation in kWh/m2.

        This includes applying broadband weighting to the RGB bands, multiplication
        by the steradians of each patch, and multiplying by the duration of time that
        they sky matrix represents in hours.

        Args:
            patch_row_str: Text string for a single row of RGB patch data.
            row_number: Integer for the row number that the patch corresponds to.
            sky_density: Integer (either 1 or 2) for the density.
            wea_duration: Number for the duration of the Wea in hours. This is used
                to convert between the average value output by the command and the
                cumulative value that is needed for all ladybug analyses.
        """
        R, G, B = patch_row_str.split(self.SEPARATOR)
        w_val = 0.265074126 * float(R) + 0.670114631 * float(G) + 0.064811243 * float(B)
        coeff = self.PATCH_ROW_COEFF[sky_density][row_number]
        return w_val * coeff * wea_duration / 1000

    def ToString(self):
        """Overwrite .NET ToString."""
        return self.__repr__()

    def __len__(self):
        if self._direct_values is None:
            self.compute_sky()
        return len(self._direct_values)

    def __getitem__(self, key):
        if self._direct_values is None:
            self.compute_sky()
        return self._direct_values[key], self._diffuse_values[key]

    def __iter__(self):
        if self._direct_values is None:
            self.compute_sky()
        return zip(self._direct_values, self._diffuse_values)

    def __repr__(self):
        """Sky Matrix object representation."""
        return "SkyMatrix [%s]" % self.wea.location.city



"""Function to compute the intersection between vectors and points."""
from __future__ import division
import os
import subprocess
import tempfile
import math

try:  # first, assume we are in cPython and numpy is installed
    from typing import Tuple
    import numpy as np
except Exception:  # we are in IronPython or numpy is not installed
    np, Tuple = None, None

from ladybug_geometry.geometry3d import Face3D, Mesh3D
from ladybug.futil import write_to_file_by_name
from ladybug.viewsphere import view_sphere

from .config import folders

if folders.radbin_path is not None:
    OCONV_EXE = os.path.join(folders.radbin_path, 'oconv.exe') if \
        os.name == 'nt' else os.path.join(folders.radbin_path, 'oconv')
    RCONTRIB_EXE = os.path.join(folders.radbin_path, 'rcontrib.exe') if \
        os.name == 'nt' else os.path.join(folders.radbin_path, 'rcontrib')
else:
    OCONV_EXE, RCONTRIB_EXE = None, None
BLACK = 'void plastic black 0 0 5 0.0 0.0 0.0 0.0 0.0'


def intersection_matrix(vectors, points, normals, context_geometry,
                        offset_distance=0, numericalize=False, sim_folder=None):
    """Compute the intersection matrix between vectors and points.

    Args:
        vectors: A list of ladybug geometry Vector3D which will be projected from
            the points of the study_mesh through the context_geometry.
        points: A list of ladybug geometry Point3D representing the sensors
            to be studied.
        normals: A list of ladybug geometry Vector3D that matches the length of
            the points list and indicate the normal direction of each point.
            THESE VECTORS MUST BE NORMALIZED.
        context_geometry: A list of ladybug geometry Face3D and/or Mesh3D that
            can block the vectors projected from the test points.
        offset_distance: An optional number to offset the sensor points before
            the vectors are cast through the context_geometry. (Default: 0).
        numericalize: A boolean to note whether the output matrix should contain
            numbers representing the cosine of the angle between the normal and
            each vector or it should simply be a matrix of booleans for whether
            the vector is seen by the surface (True) or not (False). Numbers
            can be useful for computing radiation and irradiance while the booleans
            are more helpful for direct sun and view studies. (Default: False).
        sim_folder: An optional path to a folder where the simulation files
            will be written. If None, a temporary directory will be
            used. (Default: None).

    Returns:
        A lists of lists, which can be used to account for context shade surrounding
        visualizations or geometry. The matrix will have a length equal to the points
        (and normals). Each sub-list consists of booleans and has a length equal to
        the number of vectors. True indicates that a certain patch is seen and False
        indicates that the match is blocked.
    """
    # set a default sim folder if None is specified
    if sim_folder is None:
        sim_folder = tempfile.gettempdir()
    # account for the offset distance
    if offset_distance != 0:
        points = [pt.move(vec * offset_distance) for pt, vec in zip(points, normals)]

    # write the geometry to .rad files
    geo_strs = [BLACK]
    base_geo = 'black polygon {} 0 0 {} {}'
    for i, geo in enumerate(context_geometry):
        if isinstance(geo, Face3D):
            coords = tuple(str(v) for pt in geo.vertices for v in pt.to_array())
            poly_id = 'poly_{}'.format(i)
            geo_str = base_geo.format(poly_id, len(coords), ' '.join(coords))
            geo_strs.append(geo_str)
        elif isinstance(geo, Mesh3D):
            for fi, f_geo in enumerate(geo.face_vertices):
                coords = tuple(str(v) for pt in f_geo for v in pt.to_array())
                poly_id = 'poly_{}_{}'.format(i, fi)
                geo_str = base_geo.format(poly_id, len(coords), ' '.join(coords))
                geo_strs.append(geo_str)
    scene_file = 'geometry.rad'
    write_to_file_by_name(sim_folder, scene_file, '\n'.join(geo_strs))

    # write the vectors to a file
    vec_mod = 'void light {} 0 0 3 1.0 1.0 1.0'
    base_vec = '{} source {} 0 0 4 {} 0.533'
    vec_strs, vec_mods = [], []
    for i, vec in enumerate(vectors):
        mod_id = 'vec_light{}'.format(i)
        source_id = 'vec_{}'.format(i)
        vec_mods.append(mod_id)
        vec_strs.append(vec_mod.format(mod_id))
        dir_coords = ' '.join(str(v) for v in vec.to_array())
        vec_str = base_vec.format(mod_id, source_id, dir_coords)
        vec_strs.append(vec_str)
    vec_file, vec_mod_file = 'vectors.rad', 'vectors.mod'
    write_to_file_by_name(sim_folder, vec_file, '\n'.join(vec_strs))
    write_to_file_by_name(sim_folder, vec_mod_file, '\n'.join(vec_mods))

    # create the .pts file
    sensors = []
    for pt, vec in zip(points, normals):
        sen_str = '%s %s' % (' '.join(str(v) for v in pt), ' '.join(str(v) for v in vec))
        sensors.append(sen_str)
    pts_file = 'sensors.pts'
    write_to_file_by_name(sim_folder, pts_file, '\n'.join(sensors))

    # get the environment variables and update the current working directory
    assert OCONV_EXE is not None, 'No Radiance installation was found.'
    g_env = os.environ.copy()
    if folders.env:
        for k, v in folders.env.items():
            if k.strip().upper() == 'PATH':
                g_env['PATH'] = os.pathsep.join((v, g_env['PATH']))
            else:
                g_env[k] = v
    cur_dir = os.getcwd()
    os.chdir(sim_folder)

    # create the octree
    scene_oct = 'scene.oct'
    cmd = '"{}" "{}" "{}" > "{}"'.format(OCONV_EXE, vec_file, scene_file, scene_oct)
    cmd = cmd.replace('\\', '/')
    process = subprocess.Popen(cmd, stdout=subprocess.PIPE, shell=True, env=g_env)
    process.communicate()

    # run the ray tracing command
    output_mtx = 'results.mtx'
    rad_par = '-V- -aa 0.0 -y {} -I -faf -ab 0 -dc 1.0 -dt 0.0 -dj 0.0 -dr 0 -M "{}"'
    rc_options = rad_par.format(
        len(points), os.path.join(os.path.abspath(sim_folder), vec_mod_file))
    if np is None:  # use Radiance to perform conversions on text files
        cmd = '"{}" {} "{}" < "{}"'.format(RCONTRIB_EXE, rc_options, scene_oct, pts_file)
        cmd = '{} | rmtxop -fa - -c 14713 0 0 | getinfo -  > {}'.format(cmd, output_mtx)
    else:  # leave files as binary so they can be processed with numpy
        cmd = '"{}" {} "{}" < "{}" > "{}"'.format(
            RCONTRIB_EXE, rc_options, scene_oct, pts_file, output_mtx)
    cmd = cmd.replace('\\', '/')
    process = subprocess.Popen(cmd, stdout=subprocess.PIPE, shell=True, env=g_env)
    process.communicate()

    # put back the current working directory and load the intersection matrix
    os.chdir(cur_dir)
    if np is None:  # use text parsing methods as we're in IronPython
        int_mtx = []
        with open(os.path.join(sim_folder, output_mtx), 'r') as rf:
            if numericalize:
                for row in rf:
                    int_mtx.append([float(v) for v in row.split()])
            else:
                for row in rf:
                    int_mtx.append([bool(float(v)) for v in row.split()])
    else:  # we are in cPython and we should use more efficient numpy methods
        int_mtx = binary_to_array(os.path.join(sim_folder, output_mtx))
        conversion = np.array([14713, 0, 0])
        int_mtx = np.dot(int_mtx, conversion)
        if not numericalize:
            int_mtx = int_mtx.astype(dtype=bool)

    return int_mtx


def sky_intersection_matrix(sky_matrix, points, normals, context_geometry,
                            offset_distance=0, numericalize=False, sim_folder=None):
    """Compute the intersection matrix between a sky matrix through points.

    Args:
        sky_matrix: A SkyMatrix object, which provides the vectors to be projected
            from the points through the context_geometry.
        points: A list of ladybug geometry Point3D representing the sensors
            to be studied.
        normals: A list of ladybug geometry Vector3D that matches the length of
            the points list and indicate the normal direction of each point.
        context_geometry: A list of ladybug geometry Face3D that can block the
            vectors projected from the test points.
        offset_distance: An optional number to offset the sensor points before
            the vectors are cast through the context_geometry. (Default: 0).
        numericalize: A boolean to note whether the output matrix should contain
            numbers representing the cosine of the angle between the normal and
            each vector or it should simply be a matrix of booleans for whether
            the vector is seen by the surface (True) or not (False). Numbers
            can be useful for computing radiation and irradiance while the booleans
            are more helpful for direct sun and view studies. (Default: False).
        sim_folder: An optional path to a folder where the simulation files
            will be written. If None, a temporary directory will be
            used. (Default: None).

    Returns:
        A lists of lists, which can be used to account for context shade surrounding
        visualizations or geometry. The matrix will have a length equal to the points
        (and normals). Each sub-list consists of booleans and has a length equal
        to the number of sky patches times 2 (indicating sky patches and ground patches).
        True indicates that a certain patch is seen and False indicates that the
        match is blocked.
    """
    # process the sky into an acceptable format
    lb_vecs = view_sphere.reinhart_dome_vectors if sky_matrix.high_density \
        else view_sphere.tregenza_dome_vectors
    if sky_matrix.north != 0:
        north_angle = math.radians(sky_matrix.north)
        lb_vecs = tuple(vec.rotate_xy(north_angle) for vec in lb_vecs)
    lb_grnd_vecs = tuple(vec.reverse() for vec in lb_vecs)
    vectors = lb_vecs + lb_grnd_vecs
    # compute the intersection matrix
    return intersection_matrix(
        vectors, points, normals, context_geometry,
        offset_distance, numericalize, sim_folder)


def binary_to_array(binary_file, nrows=None, ncols= None, ncomp=None, line_count=0):
    """Read a Radiance binary file as a NumPy array.

    Args:
        binary_file: Path to binary Radiance file.
        nrows: Number of rows in the Radiance file.
        ncols: Number of columns in the Radiance file.
        ncomp: Number of components of each element in the Radiance file.
        line_count: Number of lines to skip in the input file. Usually used to
            skip the header.

    Returns:
        A NumPy array.
    """
    with open(binary_file, 'rb') as reader:
        if (nrows or ncols or ncomp) is None:
            # get nrows, ncols and header line count
            nrows, ncols, ncomp, line_count = binary_mtx_dimension(binary_file)
        # skip first n lines from reader
        for i in range(line_count):
            reader.readline()

        array = np.fromfile(reader, dtype=np.float32)
        if ncomp != 1:
            array = array.reshape(nrows, ncols, ncomp)
        else:
            array = array.reshape(nrows, ncols)

    return array


def binary_mtx_dimension(filepath):
    """Return binary Radiance matrix dimensions if it exists.

    This function returns NROWS, NCOLS, NCOMP and number of header lines including the
    empty line after last header line.

    Args:
        filepath: Full path to Radiance file.

    Returns:
        A tuple with 4 integers. nrows, ncols, ncomp, line_count
    """
    try:
        inf = open(filepath, 'rb', encoding='utf-8')
    except Exception:
        inf = open(filepath, 'rb')
    try:
        first_line = next(inf).rstrip().decode('utf-8')
        if first_line[:10] != '#?RADIANCE':
            error_message = \
                'File with Radiance header must start with #?RADIANCE\n' \
                'Not {}.'.format(first_line)
            raise ValueError(error_message)

        header_lines = [first_line]
        nrows = ncols = ncomp = None
        for line in inf:
            line = line.rstrip().decode('utf-8')
            header_lines.append(line)
            if line[:6] == 'NROWS=':
                nrows = int(line.split('=')[-1])
            if line[:6] == 'NCOLS=':
                ncols = int(line.split('=')[-1])
            if line[:6] == 'NCOMP=':
                ncomp = int(line.split('=')[-1])
            if line[:7] == 'FORMAT=':
                break

        if not nrows or not ncols:
            error_message = \
                'NROWS or NCOLS was not found in the Radiance header.\nNROWS ' \
                'is {} and NCOLS is {}.\nThe header must have both ' \
                'elements.'.format(nrows, ncols)
            raise ValueError(error_message)
        return nrows, ncols, ncomp, len(header_lines) + 1
    finally:
        inf.close()



"""Class for visualizing the direct sun hours falling onto a mesh."""
from __future__ import division

try:  # first, assume we are in cPython and numpy is installed
    from typing import Tuple
    import numpy as np
except Exception:  # we are in IronPython or numpy is not installed
    np, Tuple = None, None

from ladybug_geometry.bounding import bounding_box
from ladybug_geometry.geometry3d import Vector3D, Mesh3D, Face3D
from ladybug.datatype.time import Time
from ladybug.graphic import GraphicContainer
from ladybug.legend import LegendParameters
from ladybug.color import Colorset

from ..intersection import intersection_matrix


class DirectSunStudy(object):
    """Visualize the direct sun hours falling onto a study_mesh.

    Such direct sun calculations can be used for shadow studies of outdoor
    environments or can be used to estimate glare potential from direct sun
    on the indoors.

    Args:
        vectors: Sun vectors, which will be used to determine the number of hours
            of direct sun received by the test study_mesh. These vectors can be
            computed using the ladybug.sunpath module.
        study_mesh: A ladybug geometry Mesh3D, which represents the geometry on
            which the direct sun is being studied.
        context_geometry: A list of ladybug geometry Face3D and/or Mesh3D that
            can block the sun.
        timestep: A positive integer for the number of timesteps per hour at
            which the sun vectors were generated. This is used to correctly
            interpret the time duration represented by each of the input sun
            vectors. (Default: 1 for 1 vector per hour).
        offset_distance: An optional number to offset the sensor points before
            the vectors are cast through the context_geometry. (Default: 0).
        by_vertex: A boolean to indicate whether the study should be run for
            each vertex of the study_mesh (True) or each face of the study
            mesh (False). (Default: False).
        sim_folder: An optional path to a folder where the simulation files will
            be written. If None, a temporary directory will be used. (Default: None).

    Properties:
        * vectors
        * timestep
        * study_mesh
        * context_geometry
        * offset_distance
        * by_vertex
        * sim_folder
        * study_points
        * study_normals
        * intersection_matrix
        * direct_sun_hours
    """
    __slots__ = (
        '_vectors', '_timestep', '_study_mesh', '_context_geometry',
        '_offset_distance', '_by_vertex', '_study_points', '_study_normals',
        '_sim_folder', '_intersection_matrix', '_direct_sun_hours')

    def __init__(self, vectors, study_mesh, context_geometry, timestep=1,
                 offset_distance=0, by_vertex=False, sim_folder=None):
        """Initialize RadiationDome."""
        # set default values, which will be overwritten when the study is run
        self._offset_distance = float(offset_distance)
        self._by_vertex = bool(by_vertex)
        # set the key properties of the object
        self.vectors = vectors
        self.timestep = timestep
        self.study_mesh = study_mesh
        self.context_geometry = context_geometry
        self.sim_folder = sim_folder
        # set default values, which will be overwritten when the study is run
        self._intersection_matrix = None
        self._direct_sun_hours = None

    @property
    def vectors(self):
        """Get or set a list of vectors for the sun vectors used in the study."""
        return self._vectors

    @vectors.setter
    def vectors(self, value):
        if not isinstance(value, tuple):
            try:
                value = tuple(value)
            except (ValueError, TypeError):
                raise ValueError('Expected tuple or list for vectors. '
                                 'Got {}.'.format(type(value)))
        for geo in value:
            assert isinstance(geo, Vector3D), 'Expected Vector3D for ' \
                'DirectSunStudy vectors. Got {}.'.format(type(geo))
        self._vectors = value
        self._intersection_matrix = None
        self._direct_sun_hours = None

    @property
    def timestep(self):
        """Get or set an integer for the timestep at which the vectors were generated."""
        return self._timestep

    @timestep.setter
    def timestep(self, value):
        self._timestep = int(value)
        assert self._timestep > 0, 'DirectSunStudy timestep must be ' \
            'greater than 0. Got {}.'.format(value)
        self._direct_sun_hours = None

    @property
    def study_mesh(self):
        """Get or set a SkyMatrix object for the sky used in the study."""
        return self._study_mesh

    @study_mesh.setter
    def study_mesh(self, value):
        assert isinstance(value, Mesh3D), 'Expected Mesh3D for DirectSunStudy ' \
            'study_mesh. Got {}.'.format(type(value))
        self._study_mesh = value
        self._reset_points()

    @property
    def context_geometry(self):
        """Get or set a tuple of Face3D and Mesh3D for the geometry that can block sun.
        """
        return self._context_geometry

    @context_geometry.setter
    def context_geometry(self, value):
        if not isinstance(value, tuple):
            try:
                value = tuple(value)
            except (ValueError, TypeError):
                raise ValueError('Expected tuple or list for context_geometry. '
                                 'Got {}.'.format(type(value)))
        for geo in value:
            assert isinstance(geo, (Face3D, Mesh3D)), 'Expected Face3D or Mesh3D for ' \
                'DirectSunStudy context_geometry. Got {}.'.format(type(geo))
        self._context_geometry = value
        self._intersection_matrix = None
        self._direct_sun_hours = None

    @property
    def offset_distance(self):
        """Get or set a number for the offset distance used in the study."""
        return self._offset_distance

    @offset_distance.setter
    def offset_distance(self, value):
        assert isinstance(value, (float, int)), \
            'Expected number for offset_distance. Got {}.'.format(type(value))
        self._offset_distance = value
        self._reset_points()

    @property
    def by_vertex(self):
        """Get or set a boolean for whether the study should be run for each vertex."""
        return self._by_vertex

    @by_vertex.setter
    def by_vertex(self, value):
        self._by_vertex = bool(value)
        self._reset_points()

    @property
    def sim_folder(self):
        """Get or set text for the path where the simulation files are written."""
        return self._sim_folder

    @sim_folder.setter
    def sim_folder(self, value):
        if value is not None:
            assert isinstance(value, str), 'Expected file path string for sim_folder. ' \
                'Got {}.'.format(type(value))
        self._sim_folder = value

    @property
    def study_points(self):
        """Get a tuple of Point3Ds for the points of the study."""
        return self._study_points

    @property
    def study_normals(self):
        """Get a tuple of Vector3Ds for the normals of the study."""
        return self._study_normals

    @property
    def intersection_matrix(self):
        """Get a list of lists for the intersection matrix computed by the study."""
        if self._intersection_matrix is None:
            self._compute_intersection_matrix()
        return self._intersection_matrix

    @property
    def direct_sun_hours(self):
        """Get a list of values for the number of hours falling on the study_mesh."""
        if self._direct_sun_hours is None:
            self.compute()
        return self._direct_sun_hours

    def compute(self):
        """Compute the direct sun hour values of the study.

        Note that this method is automatically called under the hood when accessing
        any results of the study and these results have not already been computed.
        So using this method is not necessary to correctly use this class. However,
        explicitly calling this method can help control when the time consuming
        part of the study runs, which is particularly helpful for larger studies.
        """
        # compute the intersection matrix
        if self._intersection_matrix is None:
            self._compute_intersection_matrix()
        # sum the intersection and sky matrices
        t_step = self.timestep
        if np is None:  # perform the calculation on float numbers
            self._direct_sun_hours = \
                [sum(int_list) / t_step for int_list in self._intersection_matrix]
        else:  # perform the calculation with numpy matrices
            self._direct_sun_hours = \
                (self._intersection_matrix.sum(axis=1) / t_step).tolist()

    def draw(self, legend_parameters=None):
        """Draw a colored study_mesh, compass, graphic/legend, and title.

        Args:
            legend_parameters: An optional LegendParameter object to change the display
                of the direct sun study. If None, default legend parameters will be
                used. (Default: None).

        Returns:
            colored_mesh: A colored Mesh3D for the study results.
            graphic: A GraphicContainer for the colored mesh, indicating the
                legend and title location for the study.
            title: Text for the title of the study.
        """
        # get the direct sun data
        d_type, unit, title = Time(), 'hr', 'Direct Sun Hours'
        sun_data = self.direct_sun_hours

        # override the legend colors sense for the direct sun study
        if legend_parameters is not None:
            assert isinstance(legend_parameters, LegendParameters), \
                'Expected LegendParameters. Got {}.'.format(type(legend_parameters))
            l_par = legend_parameters.duplicate()
        else:
            l_par = LegendParameters()
        if l_par.are_colors_default:
            l_par.colors = Colorset.ecotect()

        # create the mesh, graphic container, and title
        min_pt, max_pt = bounding_box((self.study_mesh,) + self.context_geometry)
        graphic = GraphicContainer(
            sun_data, min_pt, max_pt, l_par, d_type, unit)
        colored_mesh = self.study_mesh
        colored_mesh.colors = graphic.value_colors

        return colored_mesh, graphic, title

    def _reset_points(self):
        """Reset the study points and normals used in the study."""
        points = self._study_mesh.vertices if self._by_vertex else \
            self._study_mesh.face_centroids
        normals = self._study_mesh.vertex_normals if self._by_vertex else \
            self._study_mesh.face_normals
        if self._offset_distance != 0:
            points = tuple(
                pt.move(vec * self._offset_distance) for pt, vec in zip(points, normals))
        self._study_points = points
        self._study_normals = normals
        self._intersection_matrix = None
        self._direct_sun_hours = None

    def _compute_intersection_matrix(self):
        """Compute intersection matrix."""
        rev_vecs = [v.reverse() for v in self.vectors]
        self._intersection_matrix = intersection_matrix(
            rev_vecs, self.study_points, self.study_normals,
            self.context_geometry, self.offset_distance, numericalize=False,
            sim_folder=self.sim_folder)

    def ToString(self):
        """Overwrite .NET ToString."""
        return self.__repr__()

    def __len__(self):
        return len(self.study_points)

    def __repr__(self):
        """Direct Sun Study object representation."""
        return 'DirectSunStudy [{} values]'.format(len(self.study_points))




"""ladybug_radiance configurations.

Import this into every module where access configurations are needed.

Usage:

.. code-block:: python

    from ladybug_radiance.config import folders
    print(folders.radiance_path)
    folders.radiance_path = "C:/Radiance/bin"
"""
import ladybug.config as lb_config

import os
import platform
import subprocess
import json
import re


class Folders(object):
    """Ladybug_radiance folders.

    Args:
        config_file: The path to the config.json file from which folders are loaded.
            If None, the config.json module included in this package will be
            used. (Default: None).
        mute: If False, the paths to the various folders will be printed as they
            are found. If True, no printing will occur upon initialization of this
            class. (Default: True).

    Properties:
        * radiance_path
        * radbin_path
        * radlib_path
        * radiance_version
        * radiance_version_str
        * radiance_version_date
        * config_file
        * env
        * mute
    """

    def __init__(self, config_file=None, mute=True):
        self.mute = bool(mute)  # set the mute value
        self.config_file = config_file  # load paths from the config JSON file

    @property
    def radiance_path(self):
        """Get or set the path to Radiance installation folder.

        This is the top level folder that contains both the "bin" and the "lib"
        directories.
        """
        return self._radiance_path

    @radiance_path.setter
    def radiance_path(self, r_path):
        exe_name = 'rad.exe' if os.name == 'nt' else 'rad'
        if not r_path:  # check the default installation location
            r_path = self._find_radiance_folder()
        else:
            r_path = os.path.join(r_path, 'bin')
        rad_exe_file = os.path.join(r_path, exe_name) if r_path is not None else None

        if r_path:  # check that the Radiance executable exists in the installation
            assert os.path.isfile(rad_exe_file), \
                '{} is not a valid path to a Radiance installation.'.format(r_path)

        # set the radiance_path
        self._radbin_path = r_path
        if r_path:
            self._radiance_path = os.path.split(r_path)[0]
            radlib_path = os.path.join(self._radiance_path, 'lib')
            if os.path.isdir(radlib_path):
                self._radlib_path = radlib_path
            elif os.environ.get('RAYPATH'):
                self._radlib_path = os.environ.get('RAYPATH')
            else:
                self._radlib_path = None
            if not self.mute:
                print("Path to Radiance is set to: %s" % self._radiance_path)
        else:
            self._radiance_path = None
            self._radlib_path = None
        self._radiance_version = None
        self._radiance_version_str = None
        self._radiance_version_date = None

    @property
    def radbin_path(self):
        """Get the path to Radiance bin folder.

        This is the "bin" directory for Radiance installation (the one that
        contains the executable files).
        """
        return self._radbin_path

    @property
    def radlib_path(self):
        """Get the path to Radiance lib folder."""
        return self._radlib_path

    @property
    def radiance_version(self):
        """Get a tuple for the version of radiance (eg. (5, 3, '012cb17835')).

        This will be None if the version could not be sensed or if no Radiance
        installation was found.
        """
        if self._radbin_path and self._radiance_version_str is None:
            self._radiance_version_from_cli()
        return self._radiance_version

    @property
    def radiance_version_str(self):
        """Get text for the full version of radiance (eg."RADIANCE 5.3 official release").

        This will be None if the version could not be sensed or if no Radiance
        installation was found.
        """
        if self._radbin_path and self._radiance_version_str is None:
            self._radiance_version_from_cli()
        return self._radiance_version_str

    @property
    def radiance_version_date(self):
        """Get a tuple for the date of the radiance version (eg. (2020, 9, 3)).

        This will be None if the version could not be sensed or if no Radiance
        installation was found.
        """
        if self._radbin_path and self._radiance_version_str is None:
            self._radiance_version_from_cli()
        return self._radiance_version_date

    @property
    def config_file(self):
        """Get or set the path to the config.json file from which folders are loaded.

        Setting this to None will result in using the config.json module included
        in this package.
        """
        return self._config_file

    @config_file.setter
    def config_file(self, cfg):
        if cfg is None:
            cfg = os.path.join(os.path.dirname(__file__), 'config.json')
        self._load_from_file(cfg)
        self._config_file = cfg

    @property
    def env(self):
        "Return Radiance environment as a dictionary."
        env = {}
        if self.radbin_path:
            env['PATH'] = self.radbin_path.replace('\\', '/')
        if self.radlib_path:
            env['RAYPATH'] = self.radlib_path.replace('\\', '/')
        return env

    def _load_from_file(self, file_path):
        """Set all of the the properties of this object from a config JSON file.

        Args:
            file_path: Path to a JSON file containing the file paths. A sample of this
                JSON is the config.json file within this package.
        """
        # check the default file path
        assert os.path.isfile(str(file_path)), \
            ValueError('No file found at {}'.format(file_path))

        # set the default paths to be all blank
        default_path = {"radiance_path": r''}

        with open(file_path, 'r') as cfg:
            try:
                paths = json.load(cfg)
            except Exception as e:
                print('Failed to load paths from {}.\n{}'.format(file_path, e))
            else:
                for key, p in paths.items():
                    if not key.startswith('__') and p.strip():
                        default_path[key] = p.strip()

        # set path for radiance installations
        self.radiance_path = default_path["radiance_path"]

    def _radiance_version_from_cli(self):
        """Get the Radiance version properties by making a call to a Radiance command."""
        # check mkpmap version since this avoids interference with Accelerad
        # more information about this is here: https://discourse.ladybug.tools/t/
        # is-it-possible-to-get-lbt-1-2-0-working-with-accelerad/13789/9
        rad_exe = os.path.join(self.radbin_path, 'mkpmap.exe') if os.name == 'nt' \
            else os.path.join(self.radbin_path, 'mkpmap')
        if not os.path.isfile(rad_exe):  # old Radiance without the mkpmap executable
            rad_exe = os.path.join(self.radbin_path, 'rtrace.exe') if os.name == 'nt' \
                else os.path.join(self.radbin_path, 'rtrace')
        cmds = [rad_exe, '-version']
        use_shell = True if os.name == 'nt' else False
        process = subprocess.Popen(cmds, stdout=subprocess.PIPE, shell=use_shell)
        stdout = process.communicate()
        base_str = str(stdout[0]).replace("b'", '').replace(r"\r\n'", '')
        self._radiance_version_str = base_str  # set the version string
        if 'NREL' in base_str:  # parse using NREL's version conventions
            try:  # try to parse the version into a list of integers
                ver_nums = base_str.split(' ')[1].split('.')
                ver_array = []
                for i in ver_nums:
                    val = int(i) if i.isnumeric() else i
                    ver_array.append(val)
                self._radiance_version = tuple(ver_array)
            except Exception:
                pass  # failed to parse the version into values; possibly a custom build
            if self._radiance_version and self._radiance_version[:2] >= (5, 2):
                try:  # try to parse the date into a list of integers
                    date_pattern = re.compile(r'(\d*\-\d*\-\d*)')
                    ver_date = re.search(date_pattern, base_str)
                    self._radiance_version_date = \
                        tuple(int(v) for v in ver_date.group(0).split('-'))
                except Exception:
                    pass  # failed to parse the date into values; possibly a custom build
            else:
                try:  # try to parse the date into a list of integers
                    date_pattern = re.compile(r'(\d*\.\d*\.\d*\))')
                    ver_date = re.search(date_pattern, base_str)
                    self._radiance_version_date = \
                        tuple(int(v) for v in
                              ver_date.group(0).replace(')', '').split('.'))
                except Exception:
                    pass  # failed to parse the date into values; possibly a custom build
        else:
            try:  # try to parse the version into a list of integers
                ver_nums = base_str.split('(')[-1].split(')')[0].split('.')
                ver_array = []
                for i in ver_nums:
                    val = int(i) if i.isnumeric() else i
                    ver_array.append(val)
                self._radiance_version = tuple(ver_array)
            except Exception:
                pass  # failed to parse the version into values; possibly a custom build
            try:  # try to parse the date into a list of integers
                date_pattern = re.compile(r'(\d*\-\d*\-\d*)')
                ver_date = re.search(date_pattern, base_str)
                self._radiance_version_date = \
                    tuple(int(v) for v in ver_date.group(0).split('-'))
            except Exception:
                pass  # failed to parse the date into values; possibly a custom build

    @staticmethod
    def _find_radiance_folder():
        """Find the Radiance installation in its default location.

        This method will first attempt to return the path of Radiance installed in
        a Ladybug Tools directory. If this is not found, it will search for a
        standalone Radiance installation in its default location. Lastly, the
        environment variables will be checked.

        Returns:
            File directory and full path to executable in case of success.
            None, None in case of failure.
        """
        # first check if there's a version installed in the ladybug_tools folder
        lb_install = lb_config.folders.ladybug_tools_folder
        rad_path = None
        if os.path.isdir(lb_install):
            test_path = os.path.join(lb_install, 'radiance')
            rad_path = test_path if os.path.isdir(test_path) else None

        # then check for the default location where standalone Radiance is installed
        if rad_path is not None:
            pass  # we found a version of radiance in the ladybug_tools folder
        elif os.name == 'nt':  # search the C:/ drive on Windows
            test_path = 'C:\\Radiance'
            rad_path = test_path if os.path.isdir(test_path) else None
        elif platform.system() == 'Darwin':  # search usr/local and Applications on Mac
            test_path = '/usr/local/radiance'
            rad_path = test_path if os.path.isdir(test_path) else None
            if rad_path is None:
                test_path = '/Applications/radiance'
                rad_path = test_path if os.path.isdir(test_path) else None
        elif platform.system() == 'Linux':  # search the usr/local folder
            test_path = '/usr/local/radiance'
            rad_path = test_path if os.path.isdir(test_path) else None

        # check the environment variables
        if not rad_path:
            rad_bin_path = os.environ.get('BINPATH')
            if rad_bin_path is not None:
                return rad_bin_path

        if not rad_path:  # no Radiance installations were found
            return None

        # return the path to the executable
        rad_path = os.path.join(rad_path, 'bin')
        return rad_path


"""Object possessing all key folders within the configuration."""
folders = Folders(mute=True)
